from __future__ import annotations

from pathlib import Path
from typing import Iterable, Optional, Protocol

import h5py
from astropy.table import Column, Table  # type: ignore

from opencosmo.dataset.column import ColumnBuilder
from opencosmo.dataset.index import DataIndex


class OpenCosmoDataHandler(Protocol):
    """
    There are multiple modes we can imagine working with data in. For
    small data, it's totally fine (and probably preferable) to just load
    the dataset into memory immediately. For large data, we want to keep
    a handle to the file and only load requested data when it's needed.

    This class defines a protocol that handlers must implement. These
    handlers will be used by the Dataset class to handle data. Because the
    handler is separate from the dataset, we can have multiple datasets
    that use a single handler.

    The handler has a few responsibilities:

    1. It should only require a path to the data to work
    2. It should be a context manager
    3. It needs to be able to apply masks and transformations to the datahandler
    4. It should be able to return the data

    The handler is only responsible for working with the actual data. Indexes
    and metadata are handled separately.
    """

    def __init__(self, file: Path | h5py.File | dict): ...
    def __enter__(self): ...
    def __exit__(self, *exc_details): ...
    def __len__(self) -> int: ...
    def collect(
        self, columns: Iterable[str], index: DataIndex
    ) -> OpenCosmoDataHandler: ...
    def write(
        self,
        file: h5py.File,
        index: DataIndex,
        columns: Iterable[str],
        dataset_name: Optional[str] = None,
    ) -> None: ...
    def get_data(
        self,
        builders: dict[str, ColumnBuilder],
        index: DataIndex,
    ) -> Column | Table: ...



from pathlib import Path
from typing import Iterable, Optional

import h5py

import opencosmo as oc
from opencosmo import dataset as ds
from opencosmo.collection import Collection, SimulationCollection
from opencosmo.link.collection import StructureCollection


def open_simulation_files(**paths: Path) -> SimulationCollection:
    """
    Open multiple files and return a simulation collection. The data
    type of every file must be the same.

    Parameters
    ----------
    paths : str or Path
        The paths to the files to open.

    Returns
    -------
    SimulationCollection

    """
    datasets: dict[str, oc.Dataset] = {}
    for key, path in paths.items():
        dataset = oc.open(path)
        if not isinstance(dataset, oc.Dataset):
            raise ValueError("All datasets must be of the same type.")
    dtypes = set(dataset for dataset in datasets.values())
    if len(dtypes) != 1:
        raise ValueError("All datasets must be of the same type.")
    return SimulationCollection(datasets)


def open_multi_dataset_file(
    file: h5py.File, datasets: Optional[Iterable[str]]
) -> Collection | ds.Dataset:
    """
    Open a file with multiple datasets.
    """
    CollectionType = get_collection_type(file)
    return CollectionType.open(file, datasets)


def read_multi_dataset_file(
    file: h5py.File, datasets: Optional[Iterable[str]] = None
) -> Collection | ds.Dataset:
    """
    Read a file with multiple datasets.
    """
    CollectionType = get_collection_type(file)
    return CollectionType.read(file, datasets)


def get_collection_type(file: h5py.File) -> type[Collection]:
    """
    Determine the type of a single file containing multiple datasets. Currently
    we support multi_simulation, particle, and linked collections.

    multi_simulation == multiple simulations, same data types
    particle == single simulation, multiple particle species
    linked == A properties dataset, linked with other particle or profile datasets
    """
    datasets = [k for k in file.keys() if k != "header"]
    if len(datasets) == 0:
        raise ValueError("No datasets found in file.")

    if "header" not in file.keys():
        return SimulationCollection
    elif len(list(filter(lambda x: x.endswith("properties"), datasets))) >= 1:
        return StructureCollection
    else:
        raise ValueError(
            "Unknown file type. "
            "It appears to have multiple datasets, but organized incorrectly"
        )



from .collection import StructureCollection
from .handler import LinkHandler, OomLinkHandler
from .io import open_linked_file, open_linked_files

__all__ = [
    "StructureCollection",
    "LinkHandler",
    "OomLinkHandler",
    "open_linked_files",
    "open_linked_file",
]



from __future__ import annotations

from pathlib import Path

import h5py

try:
    from mpi4py import MPI

    from opencosmo.handler import MPIHandler
    from opencosmo.handler.mpi import partition
except ImportError:
    MPI = None  # type: ignore
from typing import Iterable, Optional

import opencosmo as oc
from opencosmo import collection
from opencosmo.dataset.index import ChunkedIndex, DataIndex
from opencosmo.file import FileExistance, file_reader, file_writer, resolve_path
from opencosmo.handler import InMemoryHandler, OpenCosmoDataHandler, OutOfMemoryHandler
from opencosmo.header import read_header
from opencosmo.transformations import units as u


def open(
    file: str | Path | h5py.File | h5py.Group,
    datasets: Optional[str | Iterable[str]] = None,
) -> oc.Dataset | collection.Collection:
    """
    Open a dataset or data collection from a file without reading the data into memory.

    The object returned by this function will only read data from the file
    when it is actually needed. This is useful if the file is very large
    and you only need to access a small part of it.

    If you open a file with this dataset, you should generally close it
    when you're done

    .. code-block:: python

        import opencosmo as oc
        ds = oc.open("path/to/file.hdf5")
        # do work
        ds.close()

    Alternatively you can use a context manager, which will close the file
    automatically when you are done with it.

    .. code-block:: python

        import opencosmo as oc
        with oc.open("path/to/file.hdf5") as ds:
            # do work

    Parameters
    ----------
    file : str or pathlib.Path
        The path to the file to open.
    datasets : str or list[str], optional
        If the file has multiple datasets, the name of the dataset(s) to open.
        All other datasets will be ignored. If not provided, will open all
        datasets

    Returns
    -------
    dataset : oc.Dataset or oc.Collection
        The dataset or collection opened from the file.

    """
    if not isinstance(file, h5py.File) and not isinstance(file, h5py.Group):
        path = resolve_path(file, FileExistance.MUST_EXIST)
        file_handle = h5py.File(path, "r")
    else:
        file_handle = file
    if "data" not in file_handle:
        if not isinstance(datasets, str):
            return collection.open_multi_dataset_file(file_handle, datasets)
        try:
            group = file_handle[datasets]
        except KeyError:
            raise ValueError(f"Dataset {datasets} not found in file {file}")
    else:
        group = file_handle

    header = read_header(file_handle)
    # tree = read_tree(file_handle, header)
    tree = None
    if datasets is not None and not isinstance(datasets, str):
        raise ValueError("Asked for multiple datasets, but file has only one")

    handler: OpenCosmoDataHandler
    index: DataIndex
    if MPI is not None and MPI.COMM_WORLD.Get_size() > 1:
        handler = MPIHandler(
            file_handle, group_name=datasets, tree=tree, comm=MPI.COMM_WORLD
        )
        start, size = partition(MPI.COMM_WORLD, len(handler))
        index = ChunkedIndex.single_chunk(start, size)
    else:
        handler = OutOfMemoryHandler(file_handle, group_name=datasets, tree=tree)
        index = ChunkedIndex.from_size(len(handler))
    builders, base_unit_transformations = u.get_default_unit_transformations(
        group, header
    )

    dataset = oc.Dataset(handler, header, builders, base_unit_transformations, index)
    return dataset


@file_reader
def read(
    file: h5py.File, datasets: Optional[str | Iterable[str]] = None
) -> oc.Dataset | collection.Collection:
    """
    Read a dataset from a file into memory.

    You should use this function if the data are small enough that having
    a copy of it (or a few copies of it) in memory is not a problem. For
    larger datasets, use :py:func:`opencosmo.open`.

    Note that some dataset types cannot be read, due to complexities with
    how the data is handled. Using :py:func:`opencosmo.open` is recommended
    for most use cases.

    Parameters
    ----------
    file : str or pathlib.Path
        The path to the file to read.
    datasets : str or list[str], optional
        If the file has multiple datasets, the name of the dataset(s) to read.
        All other datasets will be ignored. If not provided, will read all
        datasets

    Returns
    -------
    dataset : oc.Dataset or oc.Collection
        The dataset or collection read from the file.

    """

    if "data" not in file:
        if not isinstance(datasets, str):
            return collection.read_multi_dataset_file(file, datasets)
        try:
            group = file[datasets]
        except KeyError:
            raise ValueError(f"Dataset {datasets} not found in file {file}")
    else:
        group = file

    if datasets is not None and not isinstance(datasets, str):
        raise ValueError("Asked for multiple datasets, but file has only one")
    header = read_header(file)
    # tree = read_tree(file, header)
    tree = None
    handler = InMemoryHandler(file, tree, group_name=datasets)
    index = ChunkedIndex.from_size(len(handler))
    builders, base_unit_transformations = u.get_default_unit_transformations(
        group, header
    )

    return oc.Dataset(handler, header, builders, base_unit_transformations, index)


@file_writer
def write(file: h5py.File, dataset: oc.Dataset | collection.Collection) -> None:
    """
    Write a dataset or collection to the file at the sepecified path.

    Parameters
    ----------
    file : str or pathlib.Path
        The path to the file to write to.
    dataset : oc.Dataset
        The dataset to write.

    Raises
    ------
    FileExistsError
        If the file at the specified path already exists
    FileNotFoundError
        If the parent folder of the ouput file does not exist
    """
    dataset.write(file)



from typing import Protocol

from opencosmo.parameters import SimulationParameters


class SpatialIndex(Protocol):
    def __init__(self, simulation_parameters: SimulationParameters, max_level: int): ...



from __future__ import annotations

from functools import cached_property
from typing import Optional

import h5py
import numpy as np
from pydantic import BaseModel, ConfigDict, Field, model_validator

from opencosmo import parameters
from opencosmo.file import file_reader


@file_reader
def read_simulation_parameters(file: h5py.File) -> SimulationParameters:
    """
    Read the simulation parameters from an OpenCosmo file

    This function reads the simulation parameters from the header of an OpenCosmo file
    without reading the entire header or any data. It may be useful if you just want
    some basic information about the simulation.

    Parameters
    ----------

    """
    is_hydro = file["header/reformat_hacc/config"].attrs.get("is_hydro", None)
    if is_hydro is None:
        n_dm = file["header/simulation/parameters"].attrs.get("n_dm", 0)
        n_bar = file["header/simulation/parameters"].attrs.get("n_bar", 0)
        if n_dm > 0 and n_bar > 0:
            is_hydro = True
        elif n_dm > 0:
            is_hydro = False
        else:
            raise KeyError(
                "Could not determine if this simulation is hydro or gravity-only from "
                "the header. Are you sure it is an OpenCosmo file?"
            )

    try:
        cosmology_parameters = parameters.read_header_attributes(
            file, "simulation/cosmology", parameters.CosmologyParameters
        )
    except KeyError as e:
        raise KeyError(
            "This file does not appear to have cosmology information. "
            "Are you sure it is an OpenCosmo file?\n"
            f"Error: {e}"
        )

    if is_hydro:
        subrid_params = parameters.read_header_attributes(
            file,
            "simulation/parameters",
            SubgridParameters,
            cosmology_parameters=cosmology_parameters,
        )
        return parameters.read_header_attributes(
            file,
            "simulation/parameters",
            HydroSimulationParameters,
            subgrid_parameters=subrid_params,
            cosmology_parameters=cosmology_parameters,
        )
    else:
        return parameters.read_header_attributes(
            file,
            "simulation/parameters",
            GravityOnlySimulationParameters,
            cosmology_parameters=cosmology_parameters,
        )


def empty_string_to_none(v):
    if v == "":
        return None
    return v


class SimulationParameters(BaseModel):
    box_size: float = Field(ge=0, description="Size of the simulation box (Mpc/h)")
    z_ini: float = Field(ge=0.01, description="Initial redshift of the simulation")
    z_end: float = Field(ge=0.0, description="Final redshift of the simulation")
    n_gravity: Optional[int] = Field(
        ge=2,
        description="Number of gravity-only particles (per dimension). "
        'In hydrodynamic simulations, this parameter will be replaced with "n_dm"',
    )
    n_steps: int = Field(ge=1, description="Number of time steps")
    pm_grid: int = Field(ge=2, description="Number of grid points (per dimension)")
    offset_gravity_ini: Optional[float] = Field(
        description="Lagrangian offset for gravity-only particles"
    )
    cosmology_parameters: parameters.CosmologyParameters = Field(
        description="Cosmology parameters",
        exclude=True,
    )

    @model_validator(mode="before")
    @classmethod
    def empty_string_to_none(cls, data):
        if isinstance(data, dict):
            return {k: empty_string_to_none(v) for k, v in data.items()}
        return data

    @cached_property
    def step_zs(self) -> list[float]:
        """
        Get the redshift of the steps in this simulation. Outputs such that
        redshift[step_number] returns the redshift for that step. Keep in
        mind that steps go from high z -> low z.
        """
        a_ini = 1 / (1 + self.z_ini)
        a_end = 1 / (1 + self.z_end)
        # Steps are evenly spaced in log(a)
        step_as = np.logspace(np.log10(a_ini), np.log10(a_end), self.n_steps)
        return np.round(1 / step_as - 1, 3).tolist()  # type: ignore


GravityOnlySimulationParameters = SimulationParameters


def subgrid_alias_generator(name: str) -> str:
    return f"subgrid_{name}"


class SubgridParameters(BaseModel):
    model_config = ConfigDict(alias_generator=subgrid_alias_generator)
    agn_kinetic_eps: float = Field(description="AGN feedback efficiency")
    agn_kinetic_jet_vel: float = Field(description="AGN feedback velocity")
    agn_nperh: float = Field(description="AGN sphere of influence")
    agn_seed_mass: float = Field(description="AGN seed mass (Msun / h)")
    wind_egy_w: float = Field(description="Wind mass loading factor")
    wind_kappa_w: float = Field(description="Wind velocity")


class HydroSimulationParameters(SimulationParameters):
    n_gas: int = Field(
        description="Number of gas particles (per dimension)", alias="n_bar"
    )
    n_dm: int = Field(
        ge=2, description="Number of dark matter particles (per dimension)"
    )
    offset_gas_ini: float = Field(
        description="Lagrangian offset for gas particles", alias="offset_bar_ini"
    )
    offset_dm_ini: float = Field(
        description="Lagrangian offset for dark matter particles"
    )
    subgrid_parameters: SubgridParameters = Field(
        description="Parameters for subgrid hydrodynamic physics",
        exclude=True,
    )



import numpy as np
from astropy.table import Table  # type: ignore

from opencosmo.transformations import transformation as t

"""
Routines for applying transformations tables. The table that is
passed into these routines is always a copy of the raw data, so 
it is safe to modify in-place. However they should still return
the updated versions at the end.
"""


def apply_column_transformations(
    table: Table, transformations: list[t.ColumnTransformation]
):
    """
    Apply a list of column transformations to a table. If multiple
    transformations are present for the same column, they will simply
    be applied in the order they appear in the list.
    """
    for tr in transformations:
        column_name = tr.column_name
        if column_name not in table.columns:
            raise ValueError(f"Column {column_name} not found in table")
        column = table[column_name]
        if (new_column := tr(column)) is not None:
            table[column_name] = new_column
    return table


def apply_all_columns_transformations(
    table: Table, transformations: list[t.ColumnTransformation]
):
    """
    Apply a list of column transformations to all columns in the table.
    This is useful for transformations that should be applied to all columns,
    such as unit conversions.
    """
    for tr in transformations:
        for column_name in table.columns:
            column = table[column_name]
            if (new_column := tr(column)) is not None:
                table[column_name] = new_column
    return table


def apply_table_transformations(
    table: Table, transformations: list[t.TableTransformation]
):
    """
    Apply transformations to the table as a whole. These transformations
    are applied after individual column transformations.
    """
    output_table = table
    for tr in transformations:
        if (new_table := tr(output_table)) is not None:
            output_table = new_table
    return output_table


def apply_filter_transformations(
    table: Table, transformations: list[t.FilterTransformation]
):
    """
    Filter transformations produce boolean masks that are applied to the table
    to remove rows.
    """
    mask = np.ones(len(table), dtype=bool)
    for tr in transformations:
        if (new_mask := tr(table)) is not None:
            mask &= new_mask
    return table[mask]



from .collection import Collection, SimulationCollection
from .io import open_multi_dataset_file, open_simulation_files, read_multi_dataset_file

__all__ = [
    "Collection",
    "open_multi_dataset_file",
    "read_multi_dataset_file",
    "SimulationCollection",
    "open_simulation_files",
]



from .cosmology import CosmologyParameters
from .cosmotools import CosmoToolsParameters
from .file import FileParameters
from .parameters import read_header_attributes, write_header_attributes
from .reformat import ReformatParamters
from .simulation import (
    GravityOnlySimulationParameters,
    HydroSimulationParameters,
    SimulationParameters,
    SubgridParameters,
    read_simulation_parameters,
)

__all__ = [
    "FileParameters",
    "CosmologyParameters",
    "CosmoToolsParameters",
    "GravityOnlySimulationParameters",
    "HydroSimulationParameters",
    "ReformatParamters",
    "SimulationParameters",
    "SubgridParameters",
    "read_header_attributes",
    "write_header_attributes",
    "read_simulation_parameters",
]



from __future__ import annotations

import operator as op
from collections import defaultdict
from numbers import Real
from typing import Callable, Iterable

import astropy.units as u  # type: ignore
import numpy as np
from astropy import table  # type: ignore

from opencosmo.dataset.column import ColumnBuilder
from opencosmo.dataset.index import DataIndex
from opencosmo.handler import OpenCosmoDataHandler

Comparison = Callable[[float, float], bool]


def col(column_name: str) -> Column:
    return Column(column_name)


def apply_masks(
    handler: OpenCosmoDataHandler,
    column_builders: dict[str, ColumnBuilder],
    masks: Iterable[Mask],
    index: DataIndex,
) -> DataIndex:
    masks_by_column = defaultdict(list)
    for f in masks:
        masks_by_column[f.column_name].append(f)

    column_names = set(column_builders.keys())
    mask_column_names = set(masks_by_column.keys())
    if not mask_column_names.issubset(column_names):
        raise ValueError(
            "masks were applied to columns that do not exist in the dataset: "
            f"{mask_column_names - column_names}"
        )
    output_index = index

    for column_name, column_masks in masks_by_column.items():
        column_mask = np.ones(len(output_index), dtype=bool)
        builder = column_builders[column_name]
        column = handler.get_data({column_name: builder}, output_index)
        for f in column_masks:
            column_mask &= f.apply(column)
        output_index = output_index.mask(column_mask)
    return output_index


class Column:
    """
    A column representa a column in the table. This is used first and foremost
    for masking purposes. For example, if a user has loaded a dataset they
    can mask it with

    dataset.mask(oc.Col("column_name") < 5)

    In practice, this is just a factory class that returns mask
    """

    def __init__(self, column_name: str):
        self.column_name = column_name

    # mypy doesn't reason about eq and neq correctly
    def __eq__(self, other: Real | u.Quantity) -> Mask:  # type: ignore
        return Mask(self.column_name, other, op.eq)

    def __ne__(self, other: Real | u.Quantity) -> Mask:  # type: ignore
        return Mask(self.column_name, other, op.ne)

    def __gt__(self, other: Real | u.Quantity) -> Mask:
        return Mask(self.column_name, other, op.gt)

    def __ge__(self, other: Real | u.Quantity) -> Mask:
        return Mask(self.column_name, other, op.ge)

    def __lt__(self, other: Real | u.Quantity) -> Mask:
        return Mask(self.column_name, other, op.lt)

    def __le__(self, other: Real | u.Quantity) -> Mask:
        return Mask(self.column_name, other, op.le)

    def isin(self, other: Iterable[Real | u.Quantity]) -> Mask:
        return Mask(self.column_name, other, np.isin)


class Mask:
    """
    A mask is a class that represents a mask on a column. Masks evaluate
    to t/f for every element in the given column.
    """

    def __init__(
        self,
        column_name: str,
        value: float | u.Quantity,
        operator: Callable[[table.Column, float | u.Quantity], np.ndarray],
    ):
        self.column_name = column_name
        self.value = value
        self.operator = operator

    def apply(self, column: table.Column) -> np.ndarray:
        """
        mask the dataset based on the mask.
        """
        # Astropy's errors are good enough here
        value = self.value
        if not isinstance(value, u.Quantity) and column.unit is not None:
            value *= column.unit

        # mypy can't reason about columns correctly
        return self.operator(column, value)  # type: ignore



from __future__ import annotations

from typing import Generator, Iterable, Optional

import h5py
from astropy import units  # type: ignore
from astropy.cosmology import Cosmology  # type: ignore
from astropy.table import Column, Table  # type: ignore

import opencosmo.transformations as t
import opencosmo.transformations.units as u
from opencosmo.dataset.column import ColumnBuilder, get_column_builders
from opencosmo.dataset.index import ChunkedIndex, DataIndex, EmptyMaskError
from opencosmo.dataset.mask import Mask, apply_masks
from opencosmo.handler import OpenCosmoDataHandler
from opencosmo.header import OpenCosmoHeader, write_header
from opencosmo.parameters import SimulationParameters


class Dataset:
    def __init__(
        self,
        handler: OpenCosmoDataHandler,
        header: OpenCosmoHeader,
        builders: dict[str, ColumnBuilder],
        unit_transformations: dict[t.TransformationType, list[t.Transformation]],
        index: DataIndex,
    ):
        self.__handler = handler
        self.__header = header
        self.__builders = builders
        self.__base_unit_transformations = unit_transformations
        self.__index = index

    def __repr__(self):
        """
        A basic string representation of the dataset
        """
        length = len(self)
        take_length = length if length < 10 else 10
        repr_ds = self.take(take_length)
        table_repr = repr_ds.data.__repr__()
        # remove the first line
        table_repr = table_repr[table_repr.find("\n") + 1 :]
        head = f"OpenCosmo Dataset (length={length})\n"
        cosmo_repr = f"Cosmology: {self.cosmology.__repr__()}" + "\n"
        table_head = f"First {take_length} rows:\n"
        return head + cosmo_repr + table_head + table_repr

    def __len__(self):
        return len(self.__index)

    def __enter__(self):
        # Need to write tests
        return self

    def __exit__(self, *exc_details):
        return self.__handler.__exit__(*exc_details)

    def close(self):
        return self.__handler.__exit__()

    @property
    def cosmology(self) -> Cosmology:
        """
        The cosmology of the simulation this dataset is drawn from as
        an astropy.cosmology.Cosmology object.

        Returns
        -------
        cosmology: astropy.cosmology.Cosmology
        """
        return self.__header.cosmology

    @property
    def dtype(self) -> str:
        """
        The data type of this dataset.

        Returns
        -------
        dtype: str
        """
        return self.__header.file.data_type

    @property
    def redshift(self) -> float:
        """
        The redshift slice this dataset was drawn from

        Returns
        -------
        redshift: float

        """
        return self.__header.file.redshift

    @property
    def simulation(self) -> SimulationParameters:
        """
        The parameters of the simulation this dataset is drawn
        from.

        Returns
        -------
        parameters: opencosmo.parameters.SimulationParameters
        """
        return self.__header.simulation

    @property
    def data(self) -> Table | Column:
        """
        The data in the dataset. This will be an astropy.table.Table or
        astropy.table.Column (if there is only one column selected).

        Returns
        -------
        data : astropy.table.Table or astropy.table.Column
            The data in the dataset.

        """
        # should rename this, dataset.data can get confusing
        # Also the point is that there's MORE data than just the table
        return self.__handler.get_data(builders=self.__builders, index=self.__index)

    @property
    def index(self) -> DataIndex:
        return self.__index

    def filter(self, *masks: Mask) -> Dataset:
        """
        Filter the dataset based on some criteria.

        Parameters
        ----------
        *masks : Mask
            The masks to apply to dataset, constructed with :func:`opencosmo.col`

        Returns
        -------
        dataset : Dataset
            The new dataset with the masks applied.

        Raises
        ------
        ValueError
            If the given  refers to columns that are
            not in the dataset, or the  would return zero rows.

        """

        try:
            new_index = apply_masks(
                self.__handler, self.__builders, masks, self.__index
            )
        except EmptyMaskError:
            raise ValueError("No rows matched the given filters!")

        return Dataset(
            self.__handler,
            self.__header,
            self.__builders,
            self.__base_unit_transformations,
            new_index,
        )

    def rows(self) -> Generator[dict[str, float | units.Quantity]]:
        """
        Iterate over the rows in the dataset. Yields
        for each row, with associated units. For performance it is recommended
        that you first select the columns you need to work with.

        Yields
        -------
        row : dict
            A dictionary of values for each row in the dataset with units.
        """
        max = len(self)
        chunk_ranges = [(i, min(i + 1000, max)) for i in range(0, max, 1000)]
        if len(chunk_ranges) == 0:
            chunk_ranges = [(0, 0)]
        for start, end in chunk_ranges:
            chunk = self.take_range(start, end)

            chunk_data = chunk.data
            columns = {
                k: chunk_data[k].quantity if chunk_data[k].unit else chunk_data[k]
                for k in chunk_data.keys()
            }
            for i in range(len(chunk)):
                yield {k: v[i] for k, v in columns.items()}

    def select(self, columns: str | Iterable[str]) -> Dataset:
        """
        Select a subset of columns from the dataset.

        Parameters
        ----------
        columns : str or list[str]
            The column or columns to select.

        Returns
        -------
        dataset : Dataset
            The new dataset with only the selected columns.

        Raises
        ------
        ValueError
            If any of the given columns are not in the dataset.
        """
        if isinstance(columns, str):
            columns = [columns]

        # numpy compatability
        columns = [str(col) for col in columns]

        try:
            new_builders = {col: self.__builders[col] for col in columns}
        except KeyError:
            known_columns = set(self.__builders.keys())
            unknown_columns = set(columns) - known_columns
            raise ValueError(
                "Tried to select columns that aren't in this dataset! Missing columns "
                + ", ".join(unknown_columns)
            )

        return Dataset(
            self.__handler,
            self.__header,
            new_builders,
            self.__base_unit_transformations,
            self.__index,
        )

    def take(self, n: int, at: str = "random") -> Dataset:
        """
        Take some number of rows from the dataset.

        Can take the first n rows, the last n rows, or n random rows
        depending on the value of 'at'.

        Parameters
        ----------
        n : int
            The number of rows to take.
        at : str
            Where to take the rows from. One of "start", "end", or "random".
            The default is "random".

        Returns
        -------
        dataset : Dataset
            The new dataset with only the selected rows.

        Raises
        ------
        ValueError
            If n is negative or greater than the number of rows in the dataset,
            or if 'at' is invalid.

        """
        new_index = self.__index.take(n, at)

        return Dataset(
            self.__handler,
            self.__header,
            self.__builders,
            self.__base_unit_transformations,
            new_index,
        )

    def take_range(self, start: int, end: int) -> Table:
        """
        Get a range of rows from the dataset.

        Parameters
        ----------
        start : int
            The first row to get.
        end : int
            The last row to get.

        Returns
        -------
        table : astropy.table.Table
            The table with only the rows from start to end.

        Raises
        ------
        ValueError
            If start or end are negative or greater than the length of the dataset
            or if end is greater than start.

        """
        if start < 0 or end < 0:
            raise ValueError("start and end must be positive.")
        if end < start:
            raise ValueError("end must be greater than start.")
        if end > len(self):
            raise ValueError("end must be less than the length of the dataset.")

        if start < 0 or end > len(self):
            raise ValueError("start and end must be within the bounds of the dataset.")

        new_index = self.__index.take_range(start, end)

        return Dataset(
            self.__handler,
            self.__header,
            self.__builders,
            self.__base_unit_transformations,
            new_index,
        )

    def write(
        self,
        file: h5py.File | h5py.Group,
        dataset_name: Optional[str] = None,
        with_header=True,
    ) -> None:
        """
        Write the dataset to a file. This should not be called directly for the user.
        The opencosmo.write file writer automatically handles the file context.

        Parameters
        ----------
        file : h5py.File
            The file to write to.
        dataset_name : str
            The name of the dataset in the file. The default is "data".

        """
        if not isinstance(file, (h5py.File, h5py.Group)):
            raise AttributeError(
                "Dataset.write should not be called directly, "
                "use opencosmo.write instead."
            )

        if with_header:
            write_header(file, self.__header, dataset_name)

        self.__handler.write(file, self.__index, self.__builders.keys(), dataset_name)

    def with_units(self, convention: str) -> Dataset:
        """
        Create a new dataset from this one with a different unit convention.

        Parameters
        ----------
        convention : str
            The unit convention to use. One of "physical", "comoving",
            "scalefree", or "unitless".

        Returns
        -------
        dataset : Dataset
            The new dataset with the requested unit convention.

        """
        new_transformations = u.get_unit_transition_transformations(
            convention,
            self.__base_unit_transformations,
            self.__header.cosmology,
            self.redshift,
        )
        new_builders = get_column_builders(new_transformations, self.__builders.keys())

        return Dataset(
            self.__handler,
            self.__header,
            new_builders,
            self.__base_unit_transformations,
            self.__index,
        )

    def collect(self) -> Dataset:
        """
        Given a dataset that was originally opend with opencosmo.open,
        return a dataset that is in-memory as though it was read with
        opencosmo.read.

        This is useful if you have a very large dataset on disk, and you
        want to filter it down and then close the file.

        For example:

        .. code-block:: python

            import opencosmo as oc
            with oc.open("path/to/file.hdf5") as file:
                ds = file.(ds["sod_halo_mass"] > 0)
                ds = ds.select(["sod_halo_mass", "sod_halo_radius"])
                ds = ds.collect()

        The selected data will now be in memory, and the file will be closed.

        If working in an MPI context, all ranks will recieve the same data.
        """
        new_handler = self.__handler.collect(self.__builders.keys(), self.__index)
        new_index = ChunkedIndex.from_size(len(new_handler))
        return Dataset(
            new_handler,
            self.__header,
            self.__builders,
            self.__base_unit_transformations,
            new_index,
        )



from __future__ import annotations

from copy import copy
from functools import cache
from itertools import product
from typing import Iterable

from opencosmo.parameters import SimulationParameters
from opencosmo.spatial.region import Point3d

Index3d = tuple[int, int, int]


@cache
def get_index3d(p: Point3d, level: int, box_size: float) -> Index3d:
    block_size = box_size / (2**level)
    return int(p[0] // block_size), int(p[1] // block_size), int(p[2] // block_size)


def get_octtree_index(idx: Index3d, level: int, box_size: float) -> int:
    oct_idx = 0
    idx_ = copy(idx)
    for i in range(level):
        oct_idx |= (idx_[0] & 1) << 3 * i
        oct_idx |= (idx_[1] & 1) << (3 * i + 1)
        oct_idx |= (idx_[2] & 1) << (3 * i + 2)
        idx_ = (idx_[0] >> 1, idx_[1] >> 1, idx_[2] >> 1)
    return oct_idx


def get_children(idx: Index3d) -> Iterable[Index3d]:
    return (
        (idx[0] * 2 + dx, idx[1] * 2 + dy, idx[2] * 2 + dz)
        for dx, dy, dz in product(range(2), repeat=3)
    )


class OctTreeIndex:
    def __init__(self, simulation_parameters: SimulationParameters, max_level: int):
        self.simulation_parameters = simulation_parameters
        self.max_level = max_level



"""
I/O utilities for hdf5
"""

from typing import Optional

import hdf5plugin  # type: ignore
from h5py import Dataset, Group

from opencosmo.dataset.index import DataIndex


def write_index(
    input_ds: Dataset,
    output_group: Group,
    index: DataIndex,
    range_: Optional[tuple[int, int]] = None,
):
    if len(index) == 0:
        raise ValueError("No indices provided to write")
    data = index.get_data(input_ds)
    output_name = input_ds.name.split("/")[-1]
    compression = hdf5plugin.Blosc2(cname="lz4", filters=hdf5plugin.Blosc2.BITSHUFFLE)

    output_group.create_dataset(
        output_name, dtype=input_ds.dtype, data=data, compression=compression
    )
    attrs = input_ds.attrs
    for key in attrs.keys():
        output_group[output_name].attrs[key] = attrs[key]



from functools import cached_property
from typing import Optional

import h5py

from opencosmo import cosmology as cosmo
from opencosmo import parameters
from opencosmo.file import broadcast_read, file_reader, file_writer

try:
    from mpi4py import MPI
except ImportError:
    MPI = None  # type: ignore


class OpenCosmoHeader:
    """
    A class to represent the header of an OpenCosmo file. The header contains
    information about the simulation the data is a part of, as well as other
    meatadata that are useful to the library in various contexts. Most files
    will have a single unique header, but it is possible to have multiple
    headers in a SimulationCollection.
    """

    def __init__(
        self,
        file_pars: parameters.FileParameters,
        simulation_pars: parameters.SimulationParameters,
        reformat_pars: parameters.ReformatParamters,
        cosmotools_pars: parameters.CosmoToolsParameters,
    ):
        self.__file_pars = file_pars
        self.__simulation_pars = simulation_pars
        self.__reformat_pars = reformat_pars
        self.__cosmotools_pars = cosmotools_pars

    def write(self, file: h5py.File | h5py.Group) -> None:
        parameters.write_header_attributes(file, "file", self.__file_pars)

        parameters.write_header_attributes(
            file, "reformat_hacc/config", self.__reformat_pars
        )
        parameters.write_header_attributes(
            file, "simulation/parameters", self.__simulation_pars
        )
        parameters.write_header_attributes(
            file, "simulation/cosmotools", self.__cosmotools_pars
        )
        parameters.write_header_attributes(
            file, "simulation/cosmology", self.__simulation_pars.cosmology_parameters
        )
        if hasattr(self.__simulation_pars, "subgrid_parameters"):
            parameters.write_header_attributes(
                file, "simulation/parameters", self.__simulation_pars.subgrid_parameters
            )

    @cached_property
    def cosmology(self):
        return cosmo.make_cosmology(self.__simulation_pars.cosmology_parameters)

    @property
    def simulation(self):
        return self.__simulation_pars

    @property
    def reformat(self):
        return self.__reformat_pars

    @property
    def file(self):
        return self.__file_pars


@file_writer
def write_header(
    file: h5py.File, header: OpenCosmoHeader, dataset_name: Optional[str] = None
) -> None:
    """
    Write the header of an OpenCosmo file

    Parameters
    ----------
    file : h5py.File
        The file to write to
    header : OpenCosmoHeader
        The header information to write

    """
    if dataset_name is not None:
        group = file.require_group(dataset_name)
    else:
        group = file
    header.write(group)


@broadcast_read
@file_reader
def read_header(file: h5py.File | h5py.Group) -> OpenCosmoHeader:
    """
    Read the header of an OpenCosmo file

    This function may be useful if you just want to access some basic
    information about the simulation but you don't plan to actually
    read any data.

    Parameters
    ----------
    file : str | Path
        The path to the file

    Returns
    -------
    header : OpenCosmoHeader
        The header information from the file


    """
    try:
        file_parameters = parameters.read_header_attributes(
            file, "file", parameters.FileParameters
        )
    except KeyError as e:
        raise KeyError(
            "File header is malformed. Are you sure it is an OpenCosmo file?\n "
            f"Error: {e}"
        )
    try:
        reformat_parameters = parameters.read_header_attributes(
            file, "reformat_hacc/config", parameters.ReformatParamters
        )
    except KeyError as e:
        raise KeyError(
            "File header is malformed. Are you sure it is an OpenCosmo file?\n "
            f"Error: {e}"
        )
    try:
        simulation_parameters = parameters.read_simulation_parameters(file)

    except KeyError as e:
        raise KeyError(
            "This file does not appear to have simulation information. "
            "Are you sure it is an OpenCosmo file?\n"
            f"Error: {e}"
        )

    try:
        cosmotools_parameters = parameters.read_header_attributes(
            file, "simulation/cosmotools", parameters.CosmoToolsParameters
        )
    except KeyError as e:
        raise KeyError(
            "This file does not appear to have cosmotools information. "
            "Are you sure it is an OpenCosmo file?\n"
            f"Error: {e}"
        )
    return OpenCosmoHeader(
        file_parameters,
        simulation_parameters,
        reformat_parameters,
        cosmotools_parameters,
    )



# There is some weirdness that can cause things to break if this
# import is not done prior to trying to read data
import hdf5plugin  # type: ignore # noqa: F401

from .handler import OpenCosmoDataHandler
from .im import InMemoryHandler
from .oom import OutOfMemoryHandler

__all__ = ["OpenCosmoDataHandler", "InMemoryHandler", "OutOfMemoryHandler"]
try:
    import mpi4py  # noqa: F401

    from .mpi import MPIHandler  # noqa: F401

    __all__.append("MPIHandler")
except ImportError:
    pass



from datetime import date
from pathlib import Path
from typing import Optional

import numpy as np
from pydantic import BaseModel, field_serializer, field_validator, model_validator


def empty_string_to_none(value: str) -> Optional[str]:
    if type(value) is str and value == "":
        return None
    return value


class ReformatParamters(BaseModel):
    cosmotools_lc_path: Optional[Path] = None
    cosmotools_path: Path
    indat_path: Path
    is_hydro: bool
    lightcone_analysis_path_pattern: Optional[str] = None
    machine: str
    mass_threshold_sodbighaloparticles: Optional[float] = None
    mass_threshold_sodpropertybins: Optional[float] = None
    max_level: int = 0
    max_level_lc: Optional[list[tuple[int, int]]] = None
    npart_threshold_galaxyproperties: Optional[int] = None
    output_lc_path_pattern: Optional[str] = None
    rearrange_output_path_pattern: str
    rearrange_output_lc_path_pattern: Optional[str] = None
    simulation_date: date
    simulation_name: str
    snapshot_analysis_path_pattern: Optional[str] = None
    temporary_path: Optional[Path] = None

    @field_serializer(
        "cosmotools_lc_path",
        "cosmotools_path",
        "indat_path",
        "temporary_path",
        "lightcone_analysis_path_pattern",
        "output_lc_path_pattern",
        "rearrange_output_lc_path_pattern",
        "snapshot_analysis_path_pattern",
    )
    def handle_path(self, v):
        if v is None:
            return ""
        return str(v)

    @field_serializer("simulation_date")
    def handle_date(self, v):
        return v.isoformat()

    @field_validator("is_hydro", mode="before")
    @classmethod
    def numpy_bool_to_base(cls, value):
        if isinstance(value, np.bool_):
            return bool(value)
        return value

    @model_validator(mode="before")
    @classmethod
    def empty_string_to_none(cls, data):
        if isinstance(data, dict):
            data = {k: empty_string_to_none(v) for k, v in data.items()}
        return data



from __future__ import annotations

from typing import Optional

import h5py
import numpy as np

try:
    from mpi4py import MPI
except ImportError:
    MPI = None  # type: ignore

from opencosmo.header import OpenCosmoHeader
from opencosmo.spatial.index import SpatialIndex
from opencosmo.spatial.octree import OctTreeIndex


def read_tree(file: h5py.File | h5py.Group, header: OpenCosmoHeader):
    """
    Read a tree from an HDF5 file and the associated
    header. The tree is just a mapping between a spatial
    index and a slice into the data.

    Note: The max level in the header may not actually match
    the max level in the file. When a large dataset is filtered down,
    we may reduce the tree level to save space in the output file.

    The max level in the header is the maximum level in the full
    dataset, so this is the HIGHEST it can be.
    """
    max_level = header.reformat.max_level
    starts = {}
    sizes = {}

    for level in range(max_level + 1):
        try:
            group = file[f"index/level_{level}"]
        except KeyError:
            break
        level_starts = group["start"][()]
        level_sizes = group["size"][()]
        starts[level] = level_starts
        sizes[level] = level_sizes

    spatial_index = OctTreeIndex(header.simulation, max_level)
    return Tree(spatial_index, starts, sizes)


def write_tree(file: h5py.File, tree: Tree, dataset_name: str = "index"):
    tree.write(file, dataset_name)


def apply_range_mask(
    mask: np.ndarray,
    range_: tuple[int, int],
    starts: dict[int, np.ndarray],
    sizes: dict[int, np.ndarray],
) -> dict[int, tuple[int, np.ndarray]]:
    """
    Given an index range, apply a mask of the same size to produces new sizes.
    """
    output_sizes = {}

    for level, st in starts.items():
        ends = st + sizes[level]
        # Not in range if the end is less than start, or the start is greater than end
        overlaps_mask = ~((st > range_[1]) | (ends < range_[0]))
        # The first start may be less thank the range start so
        first_start_index = int(np.argmax(overlaps_mask))
        st = st[overlaps_mask]
        st[0] = range_[0]
        st = st - range_[0]
        # Determine how many true values are in the mask in the ranges
        new_sizes = np.add.reduceat(mask, st)
        output_sizes[level] = (first_start_index, new_sizes)
    return output_sizes


def pack_masked_ranges(
    old_starts: dict[int, np.ndarray],
    new_sizes: list[dict[int, tuple[int, np.ndarray]]],
    min_level_size: int = 500,
) -> tuple[dict[int, np.ndarray], dict[int, np.ndarray]]:
    """
    Given a list of masked ranges, pack them into a new set of sizes.
    This is used when working with MPI, and allows us to avoid sending
    very large masks between ranks.

    For queries that return a small fraction of the data, we can end up
    writing a lot of zeros in the lower levels of the tree. So we can
    dynamically choose to stop writing levels when the average size of
    the level is below a certain threshold
    """
    output_starts = {}
    output_sizes = {}
    for level in new_sizes[0]:
        new_level_sizes = np.zeros_like(old_starts[level])
        new_start_info = [rm[level] for rm in new_sizes]
        for first_idx, sizes in new_start_info:
            new_level_sizes[first_idx : first_idx + len(sizes)] += sizes

        avg_size = np.mean(new_level_sizes[new_level_sizes > 0])
        if avg_size < min_level_size:
            break
        output_sizes[level] = new_level_sizes
        output_starts[level] = np.cumsum(np.insert(new_level_sizes, 0, 0))[:-1]

    return output_starts, output_sizes


class Tree:
    """
    The Tree handles the spatial indexing of the data. As of right now, it's only
    functionality is to read and write the spatial index. Later we will add actual
    spatial queries
    """

    def __init__(
        self,
        index: SpatialIndex,
        starts: dict[int, np.ndarray],
        sizes: dict[int, np.ndarray],
    ):
        self.__index = index
        self.__starts = starts
        self.__sizes = sizes

    def apply_mask(
        self,
        mask: np.ndarray,
        comm: Optional[MPI.Comm] = None,
        range_: Optional[tuple] = None,
    ) -> Tree:
        """
        Given a boolean mask, create a new tree with slices adjusted to
        only include the elements where the mask is True. This is used
        when writing filtered datasets to file, or collecting.

        The mask will have the same shape as the original data.
        """

        if comm is not None and range_ is not None:
            return self.__apply_rank_mask(mask, comm, range_)
        if np.all(mask):
            return self
        output_starts = {}
        output_sizes = {}
        for level in self.__starts:
            start = self.__starts[level]
            size = self.__sizes[level]
            offsets = np.zeros_like(size)
            for i in range(len(start)):
                # Create a slice object for the current level
                s = slice(start[i], start[i] + size[i])
                slice_mask = mask[s]  # Apply the slice to the mask
                offsets[i] = np.sum(slice_mask)  # Count the number of True values
            level_starts = np.cumsum(np.insert(offsets, 0, 0))[
                :-1
            ]  # Cumulative sum to get new starts
            level_sizes = offsets
            output_starts[level] = level_starts
            output_sizes[level] = level_sizes

        return Tree(self.__index, output_starts, output_sizes)

    def __apply_rank_mask(
        self, mask: np.ndarray, comm: MPI.Comm, range_: tuple[int, int]
    ) -> Tree:
        """
        Given a range and a mask, apply the mask to the tree. The mask
        will have the same shape as the original data.
        """
        new_sizes = apply_range_mask(mask, range_, self.__starts, self.__sizes)
        all_new_sizes = comm.allgather(new_sizes)
        output_starts, output_sizes = pack_masked_ranges(self.__starts, all_new_sizes)
        return Tree(self.__index, output_starts, output_sizes)

    def write(self, file: h5py.File, dataset_name: str = "index"):
        """
        Write the tree to an HDF5 file. Note that this function
        is not responsible for applying masking. The routine calling this
        funct
        MPI = None
        MPI = None
        MPI = Noneion should first create a new tree with apply_mask if
        necessary.
        """
        group = file.require_group(dataset_name)
        for level in self.__starts:
            level_group = group.require_group(f"level_{level}")
            level_group.create_dataset("start", data=self.__starts[level])
            level_group.create_dataset("size", data=self.__sizes[level])



from __future__ import annotations

from typing import Any, Iterable, Optional

import astropy  # type: ignore
from h5py import File, Group

import opencosmo as oc
from opencosmo import link as l
from opencosmo.parameters import SimulationParameters


def filter_properties_by_dataset(
    dataset: oc.Dataset,
    properties: oc.Dataset,
    header: oc.header.OpenCosmoHeader,
    *masks,
) -> oc.Dataset:
    masked_dataset = dataset.filter(*masks)
    if header.file.data_type == "halo_properties":
        linked_column = "fof_halo_tag"
    elif header.file.data_type == "galaxy_properties":
        linked_column = "gal_tag"

    tags = masked_dataset.select(linked_column).data
    new_properties = properties.filter(oc.col(linked_column).isin(tags))
    return new_properties


class StructureCollection:
    """
    A collection of datasets that contain both high-level properties
    and lower level information (such as particles) for structures
    in the simulation. Currently these structures include halos
    and galaxies.

    For now, these are always a combination of a properties dataset
    and several particle or profile datasets.
    """

    def __init__(
        self,
        properties: oc.Dataset,
        header: oc.header.OpenCosmoHeader,
        handlers: dict[str, l.LinkHandler],
        filters: Optional[dict[str, Any]] = {},
        *args,
        **kwargs,
    ):
        """
        Initialize a linked collection with the provided datasets and links.
        """

        self.__properties = properties
        self.__header = header
        self.__handlers = handlers
        self.__index = self.__properties.index
        self.__filters = filters

    def __repr__(self):
        structure_type = self.header.file.data_type.split("_")[0] + "s"
        dtype_str = ", ".join(self.__handlers.keys())
        return f"Collection of {structure_type} with linked datasets {dtype_str}"

    def __len__(self):
        return len(self.__properties)

    @classmethod
    def open(
        cls, file: File, datasets_to_get: Optional[Iterable[str]] = None
    ) -> StructureCollection:
        return l.open_linked_file(file, datasets_to_get)

    @classmethod
    def read(cls, *args, **kwargs) -> StructureCollection:
        raise NotImplementedError

    @property
    def cosmology(self) -> astropy.cosmology.Cosmology:
        """
        The cosmology of the structure collection
        """
        return self.__properties.cosmology

    @property
    def redshift(self) -> float:
        """
        Get the redshift slice this dataset was drawn from

        Returns:
        --------
        redshift: gloat

        """
        return self.__header.file.redshift

    @property
    def simulation(self) -> SimulationParameters:
        """
        Get the parameters of the simulation this dataset is drawn
        from.

        Returns
        -------
        parameters: opencosmo.parameters.SimulationParameters
        """
        return self.__header.simulation

    @property
    def properties(self) -> oc.Dataset:
        """
        The properties dataset of the collection. Either, halo properties
        or galaxy properties.
        """
        return self.__properties

    def keys(self) -> list[str]:
        """
        Return the keys of the linked datasets.
        """
        return list(self.__handlers.keys()) + [self.__header.file.data_type]

    def values(self) -> list[oc.Dataset]:
        """
        Return the linked datasets.
        """
        return [self.__properties] + [
            handler.get_all_data() for handler in self.__handlers.values()
        ]

    def items(self) -> list[tuple[str, oc.Dataset]]:
        """
        Return the linked datasets as key-value pairs.
        """
        return [
            (key, handler.get_all_data()) for key, handler in self.__handlers.items()
        ]

    def __getitem__(self, key: str) -> oc.Dataset:
        """
        Return the linked dataset with the given key.
        """
        if key == self.__header.file.data_type:
            return self.__properties
        elif key not in self.__handlers:
            raise KeyError(f"Dataset {key} not found in collection.")
        index = self.__properties.index
        return self.__handlers[key].get_data(index)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        for dataset in self.values():
            try:
                dataset.__exit__(*args)
            except AttributeError:
                continue

    def filter(self, *masks, on_galaxies: bool = False) -> StructureCollection:
        """
        Apply a filter to the properties dataset and propagate it to the linked
        datasets. Filters are constructed with :py:func:`opencosmo.col` and behave
        exactly as they would in `opencosmo.Dataset.filter`.

        If the collection contains both halos and galaxies, the filter can be applied to
        the galaxy properties dataset by setting `on_galaxies=True`. However this will
        filter for *halos* that host galaxies that match this filter. As a result,
        galxies that do not match this filter will remain if another galaxy in their
        host halo does match.


        Parameters
        ----------
        *filters: Mask
            The filters to apply to the properties dataset constructed with
            :func:`opencosmo.col`.

        on_galaxies: bool, optional
            If True, the filter is applied to the galaxy properties dataset.

        Returns
        -------
        StructureCollection
            A new collection filtered by the given masks.

        Raises
        -------
        ValueError
            If on_galaxies is True but the collection does not contain
            a galaxy properties dataset.
        """
        if not masks:
            return self
        if not on_galaxies or self.__properties.dtype == "galaxy_properties":
            filtered = self.__properties.filter(*masks)
        elif "galaxy_properties" not in self.__handlers:
            raise ValueError("Dataset galaxy_properties not found in collection.")
        else:
            filtered = filter_properties_by_dataset(
                self["galaxy_properties"], self.__properties, self.__header, *masks
            )
        return StructureCollection(
            filtered,
            self.__header,
            self.__handlers,
        )

    def select(
        self, columns: str | Iterable[str], dataset: Optional[str] = None
    ) -> StructureCollection:
        """
        Update the linked collection to only include the columns specified
        in the given dataset. If no dataset is specified, the properties dataset
        is used.

        Parameters
        ----------
        columns : str | Iterable[str]
            The columns to select from the dataset.

        dataset : str, optional
            The dataset to select from. If None, the properties dataset is used.

        Returns
        -------
        StructureCollection
            A new collection with only the selected columns for the specified dataset.

        Raises
        -------
        ValueError
            If the specified dataset is not found in the collection.
        """
        if dataset is None or dataset == self.__header.file.data_type:
            new_properties = self.__properties.select(columns)
            return StructureCollection(
                new_properties,
                self.__header,
                self.__handlers,
            )

        elif dataset not in self.__handlers:
            raise ValueError(f"Dataset {dataset} not found in collection.")
        handler = self.__handlers[dataset]
        new_handler = handler.select(columns)
        return StructureCollection(
            self.__properties, self.__header, {**self.__handlers, dataset: new_handler}
        )

    def with_units(self, convention: str):
        """
        Apply the given unit convention to the collection.
        See :py:meth:`opencosmo.Dataset.with_units`

        Parameters
        ----------
        convention : str
            The unit convention to apply. One of "unitless", "scalefree",
            "comoving", or "physical".

        Returns
        -------
        StructureCollection
            A new collection with the unit convention applied.
        """
        new_properties = self.__properties.with_units(convention)
        new_handlers = {
            key: handler.with_units(convention)
            for key, handler in self.__handlers.items()
        }
        return StructureCollection(
            new_properties,
            self.__header,
            new_handlers,
        )

    def take(self, n: int, at: str = "random"):
        """
        Take some number of structures from the collection.
        See :py:meth:`opencosmo.Dataset.take`.

        Parameters
        ----------
        n : int
            The number of structures to take from the collection.
        at : str, optional
            The method to use to take the structures. One of "random", "first",
            or "last". Default is "random".

        Returns
        -------
        StructureCollection
            A new collection with the structures taken from the original.
        """
        new_properties = self.__properties.take(n, at)
        return StructureCollection(
            new_properties,
            self.__header,
            self.__handlers,
        )

    def objects(
        self, data_types: Optional[Iterable[str]] = None
    ) -> Iterable[tuple[dict[str, Any], oc.Dataset | dict[str, oc.Dataset]]]:
        """
        Iterate over the objects in this collection as pairs of
        (properties, datasets). For example, a halo collection could yield
        the halo properties and datasets for each of the associated partcles.

        If you don't need all the datasets, you can specify a list of data types
        for example:

        .. code-block:: python

            for row, particles in
                collection.objects(data_types=["gas_particles", "star_particles"]):
                # do work

        At each iteration, "row" will be a dictionary of halo properties with associated
        units, and "particles" will be a dictionary of datasets with the same keys as
        the data types.
        """
        if data_types is None:
            handlers = self.__handlers
        elif not all(dt in self.__handlers for dt in data_types):
            raise ValueError("Some data types are not linked in the collection.")
        else:
            handlers = {dt: self.__handlers[dt] for dt in data_types}

        for i, row in enumerate(self.__properties.rows()):
            index = self.__properties.index[i]
            output = {key: handler.get_data(index) for key, handler in handlers.items()}
            if not any(len(v) for v in output.values()):
                continue
            if len(output) == 1:
                yield row, next(iter(output.values()))
            else:
                yield row, output

    def write(self, file: File | Group):
        self.__header.write(file)
        self.__properties.write(file, self.__header.file.data_type)
        link_group = file[self.__header.file.data_type].create_group("data_linked")
        keys = list(self.__handlers.keys())
        keys.sort()
        for key in keys:
            handler = self.__handlers[key]
            handler.write(file, link_group, key, self.__index)



from typing import Iterable, Optional, Tuple
from warnings import warn

import h5py
import numpy as np
from astropy.table import Column, Table  # type: ignore
from mpi4py import MPI

from opencosmo.dataset.index import DataIndex
from opencosmo.file import get_data_structure
from opencosmo.handler import InMemoryHandler
from opencosmo.spatial.tree import Tree


def partition(comm: MPI.Comm, length: int) -> Tuple[int, int]:
    nranks = comm.Get_size()
    rank = comm.Get_rank()
    if rank == nranks - 1:
        start = rank * (length // nranks)
        size = length - start
        return (start, size)

    start = rank * (length // nranks)
    end = (rank + 1) * (length // nranks)
    size = end - start
    return (start, size)


def verify_input(comm: MPI.Comm, require: Iterable[str] = [], **kwargs) -> dict:
    """
    Verify that the input is the same on all ranks.

    If not, use the value from rank 0 if require is false,
    otherwise raise an error.
    """
    output = {}
    for key, value in kwargs.items():
        values = comm.allgather(value)

        if isinstance(value, Iterable):
            sets = [frozenset(v) for v in values]
            if len(set(sets)) > 1:
                if key in require:
                    raise ValueError(
                        f"Requested different values for {key} on different ranks."
                    )
                else:
                    warn(f"Requested different values for {key} on different ranks.")
        elif len(set(values)) > 1:
            if key in require:
                raise ValueError(
                    f"Requested different values for {key} on different ranks."
                )
            else:
                warn(f"Requested different values for {key} on different ranks.")
        output[key] = values[0]
    return output


class MPIHandler:
    """
    A handler for reading and writing data in an MPI context.
    """

    def __init__(
        self,
        file: h5py.File,
        tree: Optional[Tree] = None,
        group_name: Optional[str] = None,
        comm=MPI.COMM_WORLD,
    ):
        self.__file = file
        self.__group_name = group_name
        if group_name is None:
            self.__group = file["data"]
        else:
            self.__group = file[f"{group_name}/data"]
        self.__columns = get_data_structure(self.__group)
        self.__comm = comm
        self.__tree = tree

    def __len__(self) -> int:
        return next(iter(self.__group.values())).shape[0]

    def __enter__(self):
        return self

    def __exit__(self, *exec_details):
        self.__group = None
        self.__columns = None
        return self.__file.close()

    def collect(self, columns: Iterable[str], index: DataIndex) -> InMemoryHandler:
        # concatenate the masks from all ranks
        columns = list(columns)
        columns = verify_input(comm=self.__comm, columns=columns)["columns"]

        all_indices = self.__comm.allgather(index)
        file_path = self.__file.filename
        all_indices = all_indices[0].concatenate(*all_indices[1:])
        with h5py.File(file_path, "r") as file:
            return InMemoryHandler(
                file,
                tree=self.__tree,
                columns=columns,
                index=all_indices,
                group_name=self.__group_name,
            )

    def write(
        self,
        file: h5py.File,
        index: DataIndex,
        columns: Iterable[str],
        dataset_name: Optional[str] = None,
        selected: Optional[np.ndarray] = None,
    ) -> None:
        columns = list(columns)
        input_ = verify_input(
            comm=self.__comm,
            columns=columns,
            dataset_name=dataset_name,
            require=["dataset_name"],
        )
        columns = input_["columns"]

        # indices = redistribute_indices(indices, rank_range)

        rank_output_length = len(index)

        all_output_lengths = self.__comm.allgather(rank_output_length)

        rank = self.__comm.Get_rank()

        # Determine the number of elements this rank is responsible for
        # writing
        if not rank:
            rank_start = 0
        else:
            rank_start = np.sum(all_output_lengths[:rank])

        rank_end = rank_start + rank_output_length

        full_output_length = np.sum(all_output_lengths)
        if dataset_name is None:
            group = file
        else:
            group = file.require_group(dataset_name)
        data_group = group.create_group("data")

        for column in columns:
            # This step has to be done by all ranks, per documentation
            shape = (full_output_length,) + self.__group[column].shape[1:]
            data_group.create_dataset(column, shape, dtype=self.__group[column].dtype)
            if self.__columns[column] is not None:
                data_group[column].attrs["unit"] = self.__columns[column]

        self.__comm.Barrier()

        if rank_output_length != 0:
            for column in columns:
                data = index.get_data(self.__group[column])
                data_group[column][rank_start:rank_end] = data

        mask = np.zeros(len(self), dtype=bool)
        mask = index.set_data(mask, True)
        if self.__tree is not None:
            new_tree = self.__tree.apply_mask(mask, self.__comm, index.range())
            new_tree.write(group)  # type: ignore

        self.__comm.Barrier()

    def get_data(self, builders: dict, index: DataIndex) -> Column | Table:
        """
        Get data from the file in the range for this rank.
        """
        builder_keys = list(builders.keys())
        if self.__group is None:
            raise ValueError("This file has already been closed")

        output = {}

        for column in builder_keys:
            col = Column(index.get_data(self.__group[column]))
            output[column] = builders[column].build(col)
        if len(output) == 1:
            return next(iter(output.values()))
        return Table(output)

    def take_range(self, start: int, end: int, indices: np.ndarray) -> np.ndarray:
        if start < 0 or end > len(indices):
            raise ValueError("Requested range is not within the rank's range.")

        return indices[start:end]



from __future__ import annotations

from typing import Protocol

import numpy as np

Point3d = tuple[float, float, float]


class Region(Protocol):
    def __init__(self, *args, **kwargs): ...
    def contains(self, x: np.ndarray, y: np.ndarray, z: np.ndarray) -> np.ndarray: ...
    def bounding_box(self) -> BoxRegion: ...


class BoxRegion:
    def __init__(self, p1: Point3d, p2: Point3d):
        self.p1 = tuple([min(p1[0], p2[0]), min(p1[1], p2[1]), min(p1[2], p2[2])])
        self.p2 = tuple([max(p1[0], p2[0]), max(p1[1], p2[1]), max(p1[2], p2[2])])

    def bounding_box(self) -> BoxRegion:
        return self

    def contains(self, x: np.ndarray, y: np.ndarray, z: np.ndarray) -> np.ndarray:
        return (
            (x >= self.p1[0])
            & (x <= self.p2[0])
            & (y >= self.p1[1])
            & (y <= self.p2[1])
            & (z >= self.p1[2])
            & (z <= self.p2[2])
        )



from .collection import SimulationCollection
from .dataset import Dataset, col
from .io import open, read, write
from .link import StructureCollection, open_linked_files

__all__ = [
    "read",
    "write",
    "col",
    "open",
    "Dataset",
    "StructureCollection",
    "SimulationCollection",
    "open_linked_files",
]



from functools import partial
from typing import Optional

from astropy.table import Table  # type: ignore

from opencosmo import transformations as t


def select_columns(columns: list[str]) -> t.Transformation:
    return partial(__select_columns, columns=columns)


def __select_columns(data: Table, columns: list[str]) -> Optional[Table]:
    # astropy's error message here is not super great, so I'll make my own

    data_columns = set(data.colnames)
    invalid_columns = set(columns) - data_columns
    if invalid_columns:
        raise ValueError(
            f"Select attempted to get columns that do not exist in the dataset: "
            f"{invalid_columns}"
        )
    return data[*columns]



from __future__ import annotations

from typing import Iterable, Optional, Protocol, Self

from h5py import File, Group

import opencosmo as oc
from opencosmo.dataset.column import get_column_builders
from opencosmo.dataset.index import ChunkedIndex, DataIndex
from opencosmo.handler import OutOfMemoryHandler
from opencosmo.header import OpenCosmoHeader
from opencosmo.transformations import units as u

try:
    from mpi4py import MPI
except ImportError:
    MPI = None  # type: ignore


class DatasetBuilder(Protocol):
    """
    A DatasetBuilder is responsible for building a dataset from a file. It
    contains the logic for selecting columns and applying transformations to
    the data.
    """

    def with_units(self, convention: str) -> Self:
        pass

    def select(self, selected: str | Iterable[str]) -> Self:
        pass

    def build(
        self,
        file: File | Group,
        header: OpenCosmoHeader,
        index: Optional[DataIndex] = None,
    ) -> oc.Dataset:
        pass


class OomDatasetBuilder:
    __allowed_conventions = {
        "unitless",
        "scalefree",
        "comoving",
        "physical",
    }

    def __init__(
        self,
        selected: Optional[set[str]] = None,
        unit_convention: Optional[str] = None,
        *args,
        **kwargs,
    ):
        self.selected = selected
        self.unit_convention = (
            unit_convention if unit_convention is not None else "comoving"
        )

    def with_units(self, convention: str) -> OomDatasetBuilder:
        if convention not in self.__allowed_conventions:
            raise ValueError(
                f"Unit convention must be one of {self.__allowed_conventions}"
            )
        return OomDatasetBuilder(
            selected=self.selected,
            unit_convention=convention,
        )

    def select(self, selected: Iterable[str]) -> OomDatasetBuilder:
        selected = set(selected)
        if self.selected is None:
            return OomDatasetBuilder(
                selected=set(selected),
                unit_convention=self.unit_convention,
            )

        if not selected.issubset(self.selected):
            raise ValueError(
                "Selected columns must be a subset of the already selected columns."
            )
        return OomDatasetBuilder(
            selected=selected,
            unit_convention=self.unit_convention,
        )

    def build(
        self,
        file: File | Group,
        header: OpenCosmoHeader,
        index: Optional[DataIndex] = None,
    ) -> oc.Dataset:
        # tree = read_tree(file, header)
        tree = None
        builders, base_unit_transformations = u.get_default_unit_transformations(
            file, header
        )
        if self.selected is not None:
            selected = self.selected
        else:
            selected = builders.keys()

        if self.unit_convention != "comoving":
            new_transformations = u.get_unit_transition_transformations(
                self.unit_convention,
                base_unit_transformations,
                header.cosmology,
                header.file.redshift,
            )
            builders = get_column_builders(new_transformations, selected)

        if selected is not None:
            builders = {key: builders[key] for key in selected}

        handler = OutOfMemoryHandler(file, tree=tree)

        if index is None:
            index = ChunkedIndex.from_size(len(handler))

        dataset = oc.Dataset(
            handler,
            header,
            builders,
            base_unit_transformations,
            index,
        )
        return dataset



from __future__ import annotations

from typing import Iterable, Optional

import h5py
import numpy as np
from astropy.table import Column, Table  # type: ignore

from opencosmo.dataset.index import ChunkedIndex, DataIndex
from opencosmo.file import get_data_structure
from opencosmo.spatial.tree import Tree


class InMemoryHandler:
    """
    A handler for in-memory storage. All data will be loaded directly into memory.

    Reading a file is always done with OpenCosmo.read, which manages file opening
    and closing.
    """

    def __init__(
        self,
        file: h5py.File,
        tree: Optional[Tree] = None,
        group_name: Optional[str] = None,
        columns: Optional[Iterable[str]] = None,
        index: Optional[DataIndex] = None,
    ):
        if group_name is None:
            group = file["data"]
        else:
            group = file[f"{group_name}/data"]
        self.__columns = get_data_structure(group)
        if columns is not None:
            self.__columns = {n: u for n, u in self.__columns.items() if n in columns}
        self.__tree = tree
        if index is None:
            length = len(next(iter(group.values())))
            index = ChunkedIndex.from_size(length)
        self.__data = {
            colname: index.get_data(group[colname]) for colname in self.__columns
        }

    def __len__(self) -> int:
        return len(next(iter(self.__data.values())))

    def __enter__(self):
        return self

    def __exit__(self, *exec_details):
        return False

    def collect(self, columns: Iterable[str], index: DataIndex) -> InMemoryHandler:
        """
        Create a new InMemoryHandler with only the specified columns and
        the specified mask applied.
        """
        new_data = {
            colname: index.get_data(self.__data[colname]) for colname in columns
        }
        mask = np.zeros(len(self), dtype=bool)
        mask = index.set_data(mask, True)
        if self.__tree is not None:
            tree = self.__tree.apply_mask(mask)
        else:
            tree = None
        return InMemoryHandler(new_data, tree)

    def write(
        self,
        file: h5py.File,
        index: DataIndex,
        columns: Iterable[str],
        dataset_name: Optional[str] = None,
    ) -> None:
        """
        Write the data in the specified columns, with the specified mask, to the file.
        """
        if dataset_name is None:
            group = file
        else:
            group = file.require_group(dataset_name)
        data_group = group.require_group("data")
        for column in columns:
            data_group.create_dataset(column, data=index.get_data(self.__data[column]))
            if self.__columns[column] is not None:
                data_group[column].attrs["unit"] = self.__columns[column]
        mask = np.zeros(len(self), dtype=bool)
        if self.__tree is not None:
            tree = self.__tree.apply_mask(mask)
            tree.write(group, dataset_name="index")

    def get_data(
        self,
        builders: dict,
        index: DataIndex,
    ) -> Column | Table:
        """
        Get data from the in-memory storage with optional masking and column
        selection.
        """

        output = {}
        for column, builder in builders.items():
            col = index.get_data(self.__data[column])
            output[column] = builder.build(Column(col, name=column))

        if len(output) == 1:
            return next(iter(output.values()))
        return Table(output)



from .apply import (
    apply_all_columns_transformations,
    apply_column_transformations,
    apply_filter_transformations,
    apply_table_transformations,
)
from .generator import TransformationGenerator, generate_transformations
from .transformation import (
    AllColumnTransformation,
    ColumnTransformation,
    FilterTransformation,
    TableTransformation,
    Transformation,
    TransformationDict,
    TransformationType,
)

__all__ = [
    "AllColumnTransformation",
    "ColumnTransformation",
    "FilterTransformation",
    "TableTransformation",
    "Transformation",
    "apply_column_transformations",
    "apply_filter_transformations",
    "apply_table_transformations",
    "apply_all_columns_transformations",
    "generate_transformations",
    "TransformationGenerator",
    "TransformationDict",
    "TransformationType",
]



from __future__ import annotations

from typing import Iterable, Mapping, Optional, Protocol

try:
    from mpi4py import MPI

    from opencosmo.handler import MPIHandler
except ImportError:
    MPI = None  # type: ignore


import h5py
from astropy.cosmology import Cosmology  # type: ignore

import opencosmo as oc
from opencosmo.dataset.index import ChunkedIndex
from opencosmo.dataset.mask import Mask
from opencosmo.handler import InMemoryHandler, OpenCosmoDataHandler, OutOfMemoryHandler
from opencosmo.header import OpenCosmoHeader, read_header
from opencosmo.link import StructureCollection
from opencosmo.parameters import SimulationParameters
from opencosmo.transformations import units as u


class Collection(Protocol):
    """
    Collections represent a group of datasets that are related in some way. They
    support higher-level operations that are applied across all datasets in the
    collection, sometimes in a non-obvious way.

    This protocol defines methods a collection must implement. Most notably they
    must include  __getitem__, keys, values and __items__, which allows
    a collection to behave like a read-only dictionary.


    Note that the "open" and "read" methods are used in the case an entire collection
    is located within a single file. Multi-file collections are handled
    in the collection.io module. Most complexity is hidden from the user
    who simply calls "oc.read" and "oc.open" to get a collection. The io
    module also does sanity checking to ensure files are structurally valid,
    so we do not have to do it here.
    """

    @classmethod
    def open(
        cls, file: h5py.File, datasets_to_get: Optional[Iterable[str]] = None
    ) -> Collection | oc.Dataset: ...

    @classmethod
    def read(
        cls, file: h5py.File, datasets_to_get: Optional[Iterable[str]]
    ) -> Collection: ...

    def write(self, file: h5py.File): ...

    def __getitem__(self, key: str) -> oc.Dataset: ...
    def keys(self) -> Iterable[str]: ...
    def values(self) -> Iterable[oc.Dataset]: ...
    def items(self) -> Iterable[tuple[str, oc.Dataset]]: ...
    def __enter__(self): ...
    def __exit__(self, *exc_details): ...
    def filter(self, *masks: Mask) -> Collection: ...
    def select(self, *args, **kwargs) -> Collection: ...
    def with_units(self, convention: str) -> Collection: ...
    def take(self, *args, **kwargs) -> Collection: ...


def write_with_common_header(
    collection: Collection, header: OpenCosmoHeader, file: h5py.File
):
    """
    Write a collection to an HDF5 file when all datasets share
    a common header.
    """
    # figure out if we have unique headers

    header.write(file)
    keys = list(collection.keys())
    keys.sort()
    for key in keys:
        group = file.create_group(key)
        collection[key].write(group, key, with_header=False)


def write_with_unique_headers(collection: Collection, file: h5py.File):
    """
    Write the collection to an HDF5 file when each dattaset
    has its own header.
    """
    # figure out if we have unique headers

    keys = list(collection.keys())
    keys.sort()
    for key in keys:
        group = file.create_group(key)
        collection[key].write(group)


def verify_datasets_exist(file: h5py.File, datasets: Iterable[str]):
    """
    Verify a set of datasets exist in a given file.
    """
    if not set(datasets).issubset(set(file.keys())):
        raise ValueError(f"Some of {', '.join(datasets)} not found in file.")


class SimulationCollection(dict):
    """
    A collection of datasets of the same type from different
    simulations. In general this exposes the exact same API
    as the individual datasets, but maps the results across
    all of them.
    """

    def __init__(self, datasets: Mapping[str, oc.Dataset | Collection]):
        self.update(datasets)

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        for dataset in self.values():
            try:
                dataset.close()
            except ValueError:
                continue

    def __repr__(self):
        n_collections = sum(
            1
            for v in self.values()
            if isinstance(v, (SimulationCollection, StructureCollection))
        )
        n_datasets = sum(1 for v in self.values() if isinstance(v, oc.Dataset))
        return (
            f"SimulationCollection({n_collections} collections, {n_datasets} datasets)"
        )

    @classmethod
    def open(
        cls, file: h5py.File, datasets_to_get: Optional[Iterable[str]] = None
    ) -> SimulationCollection:
        if datasets_to_get is not None:
            verify_datasets_exist(file, datasets_to_get)
            names = datasets_to_get
        else:
            names = list(filter(lambda x: x != "header", file.keys()))
        datasets = {name: oc.open(file[name]) for name in names}
        return cls(datasets)

    @classmethod
    def read(
        cls, file: h5py.File, datasets_to_get: Optional[Iterable[str]] = None
    ) -> SimulationCollection:
        if datasets_to_get is not None:
            verify_datasets_exist(file, datasets_to_get)
            names = datasets_to_get
        else:
            names = list(filter(lambda x: x != "header", file.keys()))

        datasets = {name: read_single_dataset(file, name) for name in names}
        return cls(datasets)

    def write(self, h5file: h5py.File):
        return write_with_unique_headers(self, h5file)

    def __map(self, method, *args, **kwargs):
        """
        This type of collection will only ever be constructed if all the underlying
        datasets have the same data type, so it is always safe to map operations
        across all of them.
        """
        output = {k: getattr(v, method)(*args, **kwargs) for k, v in self.items()}
        return SimulationCollection(output)

    def __map_attribute(self, attribute):
        return {k: getattr(v, attribute) for k, v in self.items()}

    @property
    def cosmology(self) -> dict[str, Cosmology]:
        """
        Get the cosmologies of the simulations in the collection

        Returns
        --------
        cosmologies: dict[str, astropy.cosmology.Cosmology]
        """
        return self.__map_attribute("cosmology")

    @property
    def redshift(self) -> dict[str, float]:
        """
        Get the redshift slices for the simulations in the collection

        Returns
        --------
        redshifts: dict[str, float]
        """
        return self.__map_attribute("redshift")

    @property
    def simulation(self) -> dict[str, SimulationParameters]:
        """
        Get the simulation parameters for the simulations in the collection

        Returns
        --------
        simulation_parameters: dict[str, opencosmo.parameters.SimulationParameters]
        """

        return self.__map_attribute("simulation")

    def filter(self, *masks: Mask, **kwargs) -> SimulationCollection:
        """
        Filter the datasets in the collection. This method behaves
        exactly like :meth:`opencosmo.Dataset.filter` or
        :meth:`opencosmo.StructureCollection.filter`, but
        it applies the filter to all the datasets or collections
        within this collection. The result is a new collection.

        Parameters
        ----------
        filters:
            The filters constructed with :func:`opencosmo.col`

        Returns
        -------
        SimulationCollection
            A new collection with the same datasets, but only the
            particles that pass the filter.
        """
        return self.__map("filter", *masks, **kwargs)

    def select(self, *args, **kwargs) -> SimulationCollection:
        """
        Select a subset of the datasets in the collection. This method
        calls the underlying method in :class:`opencosmo.Dataset`, or
        :class:`opencosmo.Collection` depending on the context. As such
        its behavior and arguments can vary depending on what this collection
        contains.

        Parameters
        ----------
        args:
            The arguments to pass to the select method. This is
            usually a list of column names to select.
        kwargs:
            The keyword arguments to pass to the select method.
            This is usually a dictionary of column names to select.

        """
        return self.__map("select", *args, **kwargs)

    def take(self, n: int, at: str = "random") -> SimulationCollection:
        """
        Take a subest of rows from all datasets or collections in this collection.
        This method will delegate to the underlying method in
        :class:`opencosmo.Dataset`, or :class:`opencosmo.StructureCollection` depending
        on  the context. As such, behavior may vary depending on what this collection
        contains. See their documentation for more info.

        Parameters
        ----------
        n: int
            The number of rows to take
        at: str, default = "random"
            The method to use to take rows. Must be one of "start", "end", "random".

        """
        if any(len(ds) < n for ds in self.values()):
            raise ValueError(
                f"Not all datasets in this collection have at least {n} rows!"
            )
        return self.__map("take", n, at)

    def with_units(self, convention: str) -> SimulationCollection:
        """
        Transform all datasets or collections to use the given unit convention. This
        method behaves exactly like :meth:`opencosmo.Dataset.with_units`.

        Parameters
        ----------
        convention: str
            The unit convention to use. One of "unitless",
            "scalefree", "comoving", or "physical".

        """
        return self.__map("with_units", convention)


def open_single_dataset(
    file: h5py.File, dataset_key: str, header: Optional[OpenCosmoHeader] = None
) -> oc.Dataset:
    """
    Open a single dataset in a file with multiple datasets.
    """
    if dataset_key not in file.keys():
        raise ValueError(f"No group named '{dataset_key}' found in file.")

    if header is None:
        header = read_header(file[dataset_key])

    # tree = read_tree(file[dataset_key], header)
    tree = None
    handler: OpenCosmoDataHandler
    if MPI is not None and MPI.COMM_WORLD.Get_size() > 1:
        handler = MPIHandler(
            file, tree=tree, comm=MPI.COMM_WORLD, group_name=dataset_key
        )
    else:
        handler = OutOfMemoryHandler(file, tree=tree, group_name=dataset_key)

    builders, base_unit_transformations = u.get_default_unit_transformations(
        file[dataset_key], header
    )
    index = ChunkedIndex.from_size(len(handler))
    return oc.Dataset(handler, header, builders, base_unit_transformations, index)


def read_single_dataset(
    file: h5py.File, dataset_key: str, header: Optional[OpenCosmoHeader] = None
):
    """
    Read a single dataset from a multi-dataset file
    """
    if dataset_key not in file.keys():
        raise ValueError(f"No group named '{dataset_key}' found in file.")

    if header is None:
        header = read_header(file[dataset_key])

    # tree = read_tree(file[dataset_key], header)
    tree = None

    handler = InMemoryHandler(file, tree, dataset_key)
    builders, base_unit_transformations = u.get_default_unit_transformations(
        file[dataset_key], header
    )
    index = ChunkedIndex.from_size(len(handler))
    return oc.Dataset(handler, header, builders, base_unit_transformations, index)



from __future__ import annotations

from typing import Iterable, Optional

import numpy as np
from h5py import File, Group
from mpi4py import MPI

import opencosmo as oc
from opencosmo.dataset.column import get_column_builders
from opencosmo.dataset.index import ChunkedIndex, DataIndex, SimpleIndex
from opencosmo.handler import MPIHandler
from opencosmo.handler.mpi import partition
from opencosmo.header import OpenCosmoHeader
from opencosmo.link.builder import DatasetBuilder
from opencosmo.spatial import Tree
from opencosmo.transformations import units as u


class MpiLinkHandler:
    def __init__(
        self,
        file: File | Group,
        link: Group | tuple[Group, Group],
        header: OpenCosmoHeader,
        builder: Optional[DatasetBuilder] = None,
        comm: MPI.Comm = MPI.COMM_WORLD,
        **kwargs,
    ):
        self.selected: Optional[set[str]] = None
        self.file = file
        self.link = link
        self.header = header
        self.comm = comm
        if builder is None:
            # tree = read_tree(file, self.header)
            tree = None
            self.builder: DatasetBuilder = MpiDatasetBuilder(tree, comm=comm)
        else:
            self.builder = builder
        if isinstance(self.link, tuple):
            n_per_rank = self.link[0].shape[0] // self.comm.Get_size()
            self.offset = n_per_rank * self.comm.Get_rank()
        else:
            n_per_rank = self.link.shape[0] // self.comm.Get_size()
            self.offset = n_per_rank * self.comm.Get_rank()

    def get_all_data(self) -> oc.Dataset:
        return self.builder.build(
            self.file,
            self.header,
        )

    def get_data(self, index: DataIndex) -> oc.Dataset:
        if isinstance(self.link, tuple):
            start = index.get_data(self.link[0])
            size = index.get_data(self.link[1])
            valid_rows = size > 0
            start = start[valid_rows]
            size = size[valid_rows]
            new_index: DataIndex
            if len(start) == 0:
                new_index = SimpleIndex(np.array([], dtype=int))
            else:
                new_index = ChunkedIndex(start, size)
        else:
            indices_into_data = index.get_data(self.link)
            indices_into_data = indices_into_data[indices_into_data >= 0]
            if len(indices_into_data) == 0:
                indices_into_data = np.array([], dtype=int)
            new_index = SimpleIndex(indices_into_data)

        dataset = self.builder.build(
            self.file,
            self.header,
            index=new_index,
        )

        return dataset

    def with_units(self, convention: str) -> MpiLinkHandler:
        new_builder = self.builder.with_units(convention)
        return MpiLinkHandler(
            self.file,
            self.link,
            self.header,
            comm=self.comm,
            builder=new_builder,
        )

    def select(self, columns: str | Iterable[str]) -> MpiLinkHandler:
        new_builder = self.builder.select(columns)
        return MpiLinkHandler(
            self.file,
            self.link,
            self.header,
            comm=self.comm,
            builder=new_builder,
        )

    def write(
        self, data_group: Group, link_group: Group, name: str, index: DataIndex
    ) -> None:
        # Pack the indices
        sizes = self.comm.allgather(len(index))
        shape = (sum(sizes),)
        if sum(sizes) == 0:
            return

        if not isinstance(self.link, tuple):
            link_group.create_dataset("sod_profile_idx", shape=shape, dtype=int)
            self.comm.Barrier()
            indices_into_data = index.get_data(self.link)
            nonzero = indices_into_data >= 0
            all_nonzero = self.comm.gather(nonzero)

            if self.comm.Get_rank() == 0:
                if all_nonzero is None:
                    # should never happen, but mypy...
                    raise ValueError("No data to write")
                nonzero = np.concatenate(all_nonzero)
                sod_profile_idx = np.full(len(nonzero), -1)
                sod_profile_idx[nonzero] = np.arange(sum(nonzero))
                link_group["sod_profile_idx"][:] = sod_profile_idx

        else:
            link_group.create_dataset(f"{name}_start", shape=shape, dtype=int)
            link_group.create_dataset(f"{name}_size", shape=shape, dtype=int)
            self.comm.Barrier()
            rank_sizes = index.get_data(self.link[1])
            all_rank_sizes = self.comm.gather(rank_sizes)
            if self.comm.Get_rank() == 0:
                if all_rank_sizes is None:
                    # should never happen, but mypy...
                    raise ValueError("No data to write")

                all_sizes = np.concatenate(all_rank_sizes)
                starts = np.insert(np.cumsum(all_sizes), 0, 0)[:-1]
                link_group[f"{name}_start"][:] = starts
                link_group[f"{name}_size"][:] = all_sizes

        self.comm.Barrier()
        dataset = self.get_data(index)
        if dataset is not None:
            dataset.write(data_group, name)


class MpiDatasetBuilder:
    __allowed_conventions = {
        "unitless",
        "scalefree",
        "comoving",
        "physical",
    }

    def __init__(
        self,
        tree: Optional[Tree] = None,
        selected: Optional[set[str]] = None,
        unit_convention: Optional[str] = None,
        comm: MPI.Comm = MPI.COMM_WORLD,
    ):
        self.tree = tree
        self.selected = selected
        self.unit_convention = (
            unit_convention if unit_convention is not None else "comoving"
        )
        self.comm = comm

    def with_units(self, convention: str) -> MpiDatasetBuilder:
        if convention not in self.__allowed_conventions:
            raise ValueError(
                f"Unit convention must be one of {self.__allowed_conventions}"
            )
        return MpiDatasetBuilder(
            tree=self.tree,
            selected=self.selected,
            unit_convention=convention,
            comm=self.comm,
        )

    def select(self, selected: str | Iterable[str]) -> MpiDatasetBuilder:
        if isinstance(selected, str):
            selected = [selected]
        selected = set(selected)
        if self.selected is None:
            return MpiDatasetBuilder(
                tree=self.tree,
                selected=selected,
                unit_convention=self.unit_convention,
                comm=self.comm,
            )
        if not selected.issubset(self.selected):
            raise ValueError(
                "Selected columns must be a subset of the already selected columns."
            )
        return MpiDatasetBuilder(
            tree=self.tree,
            selected=self.selected.intersection(selected),
            unit_convention=self.unit_convention,
            comm=self.comm,
        )

    def build(
        self,
        file: File | Group,
        header: OpenCosmoHeader,
        index: Optional[DataIndex] = None,
    ) -> oc.Dataset:
        builders, base_unit_transformations = u.get_default_unit_transformations(
            file, header
        )
        selected = self.selected
        if selected is None:
            selected = builders.keys()

        if self.unit_convention != "comoving":
            new_transformations = u.get_unit_transition_transformations(
                self.unit_convention,
                base_unit_transformations,
                header.cosmology,
                header.file.redshift,
            )
            builders = get_column_builders(new_transformations, selected)

        builders = {key: builders[key] for key in selected}

        handler = MPIHandler(file, tree=self.tree, comm=self.comm)
        if index is None:
            start, size = partition(self.comm, len(handler))
            index = ChunkedIndex.single_chunk(start, size)

        dataset = oc.Dataset(
            handler,
            header,
            builders,
            base_unit_transformations,
            index,
        )

        return dataset



from __future__ import annotations

from typing import Iterable, Optional

import h5py
import numpy as np
from astropy.table import Column, Table  # type: ignore

from opencosmo.dataset.index import DataIndex
from opencosmo.handler import InMemoryHandler
from opencosmo.spatial.tree import Tree
from opencosmo.utils import write_index


class OutOfMemoryHandler:
    """
    A handler for data that will not be stored in memory. Data will remain on
    disk until needed
    """

    def __init__(
        self,
        file: h5py.File,
        tree: Optional[Tree] = None,
        group_name: Optional[str] = None,
    ):
        self.__group_name = group_name
        self.__file = file
        if group_name is None:
            self.__group = file["data"]
        else:
            self.__group = file[f"{group_name}/data"]
        self.__tree = tree

    def __len__(self) -> int:
        first_column_name = next(iter(self.__group.keys()))
        return self.__group[first_column_name].shape[0]

    def __enter__(self):
        return self

    def __exit__(self, *exec_details):
        self.__group = None
        return self.__file.close()

    def collect(self, columns: Iterable[str], index: DataIndex) -> InMemoryHandler:
        file_path = self.__file.filename
        tree: Optional[Tree] = None
        if self.__tree is not None and len(index) == len(self):
            mask = np.zeros(len(self), dtype=bool)
            mask = index.set_data(mask, True)
            tree = self.__tree.apply_mask(mask)

        else:
            tree = self.__tree

        with h5py.File(file_path, "r") as file:
            return InMemoryHandler(
                file,
                tree,
                group_name=self.__group_name,
                columns=columns,
                index=index,
            )

    def write(
        self,
        file: h5py.File,
        index: DataIndex,
        columns: Iterable[str],
        dataset_name: Optional[str] = None,
    ) -> None:
        if self.__group is None:
            raise ValueError("This file has already been closed")
        if dataset_name is None:
            group = file
        else:
            group = file.require_group(dataset_name)
        data_group = group.create_group("data")
        for column in columns:
            write_index(self.__group[column], data_group, index)

        if self.__tree is not None:
            tree_mask = np.zeros(len(self), dtype=bool)
            tree_mask = index.set_data(tree_mask, True)
            tree = self.__tree.apply_mask(tree_mask)
            tree.write(group)

    def get_data(self, builders: dict, index: DataIndex) -> Column | Table:
        """ """
        if self.__group is None:
            raise ValueError("This file has already been closed")
        output = {}
        for column, builder in builders.items():
            col = Column(index.get_data(self.__group[column]))
            output[column] = builder.build(col)

        if len(output) == 1:
            return next(iter(output.values()))
        return Table(output)

    def take_range(self, start: int, end: int, indices: np.ndarray) -> np.ndarray:
        if start < 0 or end > len(indices):
            raise ValueError("Indices out of range")
        return indices[start:end]



from enum import Enum

from pydantic import BaseModel, ConfigDict, field_validator


class FileType(Enum):
    galaxy_properties = "galaxy_properties"
    galaxy_particles = "galaxy_particles"
    halo_properties = "halo_properties"
    halo_profiles = "halo_profiles"
    halo_particles = "halo_particles"


class FileParameters(BaseModel):
    model_config = ConfigDict(use_enum_values=True)
    data_type: FileType
    is_lightcone: bool
    redshift: float
    step: int

    @field_validator("is_lightcone", mode="before")
    def validate_is_lightcone(cls, value):
        return bool(value)



from __future__ import annotations

from typing import Any, Protocol, TypeGuard, TypeVar

import h5py
import numpy as np

T = TypeVar("T", np.ndarray, h5py.Dataset)


class EmptyMaskError(Exception):
    pass


def all_are_chunked(
    others: tuple[DataIndex, ...],
) -> TypeGuard[tuple[ChunkedIndex, ...]]:
    """
    Check if all elements in the tuple are instances of ChunkedIndex.
    """
    return all(isinstance(other, ChunkedIndex) for other in others)


def all_are_simple(others: tuple[DataIndex, ...]) -> TypeGuard[tuple[SimpleIndex, ...]]:
    """
    Check if all elements in the tuple are instances of SimpleIndex.
    """
    return all(isinstance(other, SimpleIndex) for other in others)


class DataIndex(Protocol):
    @classmethod
    def from_size(cls, size: int) -> DataIndex: ...
    def set_data(self, data: np.ndarray, value: Any) -> np.ndarray: ...
    def get_data(self, data: h5py.Dataset | np.ndarray) -> np.ndarray: ...
    def take(self, n: int, at: str = "random") -> DataIndex: ...
    def take_range(self, start: int, end: int) -> DataIndex: ...
    def mask(self, mask: np.ndarray) -> DataIndex: ...
    def range(self) -> tuple[int, int]: ...
    def concatenate(self, *others: DataIndex) -> DataIndex: ...
    def __len__(self) -> int: ...
    def __getitem__(self, item: int) -> DataIndex: ...


class SimpleIndex:
    """
    An index of integers.
    """

    def __init__(self, index: np.ndarray) -> None:
        self.__index = np.sort(index)

    @classmethod
    def from_size(cls, size: int) -> DataIndex:
        return SimpleIndex(np.arange(size))

    def __len__(self) -> int:
        return len(self.__index)

    def range(self) -> tuple[int, int]:
        """
        Guranteed to be sorted
        """
        return self.__index[0], self.__index[-1]

    def concatenate(self, *others: DataIndex) -> DataIndex:
        if len(others) == 0:
            return self
        if all_are_simple(others):
            new_index = np.concatenate(
                [self.__index] + [other.__index for other in others]
            )
            new_index = np.sort(np.unique(new_index))
            return SimpleIndex(new_index)
        else:
            simple_indices = map(
                lambda x: x.to_simple_index() if isinstance(x, ChunkedIndex) else x,
                others,
            )
            return self.concatenate(*simple_indices)

    def set_data(self, data: np.ndarray, value: bool) -> np.ndarray:
        """
        Set the data at the index to the given value.
        """
        if not isinstance(data, np.ndarray):
            raise ValueError("Data must be a numpy array")

        data[self.__index] = value
        return data

    def take(self, n: int, at: str = "random") -> DataIndex:
        """
        Take n elements from the index.
        """
        if n > len(self):
            raise ValueError(f"Cannot take {n} elements from index of size {len(self)}")
        if at == "random":
            return SimpleIndex(np.random.choice(self.__index, n, replace=False))
        elif at == "start":
            return SimpleIndex(self.__index[:n])
        elif at == "end":
            return SimpleIndex(self.__index[-n:])
        else:
            raise ValueError(f"Unknown value for 'at': {at}")

    def take_range(self, start: int, end: int) -> DataIndex:
        """
        Take a range of elements from the index.
        """
        if start < 0 or end > len(self):
            raise ValueError(
                f"Range {start}:{end} is out of bounds for index of size {len(self)}"
            )

        if start >= end:
            raise ValueError(f"Start {start} must be less than end {end}")

        return SimpleIndex(self.__index[start:end])

    def mask(self, mask: np.ndarray) -> DataIndex:
        if mask.shape != self.__index.shape:
            raise np.exceptions.AxisError(
                f"Mask shape {mask.shape} does not match index size {len(self)}"
            )

        if mask.dtype != bool:
            raise TypeError(f"Mask dtype {mask.dtype} is not boolean")

        if not mask.any():
            raise EmptyMaskError("Mask is all False")

        if mask.all():
            return self

        return SimpleIndex(self.__index[mask])

    def get_data(self, data: h5py.Dataset) -> np.ndarray:
        """
        Get the data from the dataset using the index.
        """
        if not isinstance(data, (h5py.Dataset, np.ndarray)):
            raise ValueError("Data must be a h5py.Dataset")
        if len(self) == 0:
            return np.array([], dtype=data.dtype)

        min_index = self.__index.min()
        max_index = self.__index.max()
        output = data[min_index : max_index + 1]
        indices_into_output = self.__index - min_index
        return output[indices_into_output]

    def __getitem__(self, item: int) -> DataIndex:
        """
        Get an item from the index.
        """
        if item < 0 or item >= len(self):
            raise IndexError(
                f"Index {item} out of bounds for index of size {len(self)}"
            )
        val = self.__index[item]
        return SimpleIndex(np.array([val]))


def pack(start: np.ndarray, size: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Combine adjacent chunks into a single chunk.
    """

    # Calculate the end of each chunk
    end = start + size

    # Determine where a new chunk should start (i.e., not adjacent to previous)
    # We prepend True for the first chunk to always start a group
    new_group = np.ones(len(start), dtype=bool)
    new_group[1:] = start[1:] != end[:-1]

    # Assign a group ID for each segment
    group_ids = np.cumsum(new_group)

    # Combine chunks by group
    combined_start = np.zeros(group_ids[-1], dtype=start.dtype)
    combined_size = np.zeros_like(combined_start)

    np.add.at(combined_start, group_ids - 1, np.where(new_group, start, 0))
    np.add.at(combined_size, group_ids - 1, size)

    return combined_start, combined_size


class ChunkedIndex:
    def __init__(self, starts: np.ndarray, sizes: np.ndarray) -> None:
        # sort the starts and sizes
        # pack the starts and sizes
        self.__starts = starts
        self.__sizes = sizes

    def range(self) -> tuple[int, int]:
        """
        Get the range of the index.
        """
        return self.__starts[0], self.__starts[-1] + self.__sizes[-1] - 1

    def to_simple_index(self) -> SimpleIndex:
        """
        Convert the ChunkedIndex to a SimpleIndex.
        """
        idxs = np.concatenate(
            [
                np.arange(start, start + size)
                for start, size in zip(self.__starts, self.__sizes)
            ]
        )
        idxs = np.unique(idxs)
        return SimpleIndex(idxs)

    def concatenate(self, *others: DataIndex) -> DataIndex:
        if len(others) == 0:
            return self
        if all_are_chunked(others):
            new_starts = np.concatenate(
                [self.__starts] + [other.__starts for other in others]
            )
            new_sizes = np.concatenate(
                [self.__sizes] + [other.__sizes for other in others]
            )
            return ChunkedIndex(new_starts, new_sizes)

        else:
            simple_indices = map(
                lambda x: x.to_simple_index() if isinstance(x, ChunkedIndex) else x,
                others,
            )
            return self.concatenate(*simple_indices)

    @classmethod
    def from_size(cls, size: int) -> DataIndex:
        """
        Create a ChunkedIndex from a size.
        """
        if size <= 0:
            raise ValueError(f"Size must be positive, got {size}")
        # Create an array of chunk sizes

        starts = np.array([0])
        sizes = np.array([size])
        return ChunkedIndex(starts, sizes)

    @classmethod
    def single_chunk(cls, start: int, size: int) -> ChunkedIndex:
        """
        Create a ChunkedIndex with a single chunk.
        """
        if size <= 0:
            raise ValueError(f"Size must be positive, got {size}")
        if start < 0:
            raise ValueError(f"Start must be non-negative, got {start}")
        starts = np.array([start])
        sizes = np.array([size])
        return ChunkedIndex(starts, sizes)

    def set_data(self, data: np.ndarray, value: bool) -> np.ndarray:
        """
        Set the data at the index to the given value.
        """
        if not isinstance(data, np.ndarray):
            raise ValueError("Data must be a numpy array")

        for start, size in zip(self.__starts, self.__sizes):
            data[start : start + size] = value
        return data

    def __len__(self) -> int:
        """
        Get the total size of the index.
        """
        return np.sum(self.__sizes)

    def take(self, n: int, at: str = "random") -> DataIndex:
        if n > len(self):
            raise ValueError(f"Cannot take {n} elements from index of size {len(self)}")

        if at == "random":
            idxs = np.concatenate(
                [
                    np.arange(start, start + size)
                    for start, size in zip(self.__starts, self.__sizes)
                ]
            )
            idxs = np.random.choice(idxs, n, replace=False)
            return SimpleIndex(idxs)

        elif at == "start":
            last_chunk_in_range = np.searchsorted(np.cumsum(self.__sizes), n)
            new_starts = self.__starts[: last_chunk_in_range + 1].copy()
            new_sizes = self.__sizes[: last_chunk_in_range + 1].copy()
            new_sizes[-1] = n - np.sum(new_sizes[:-1])
            return ChunkedIndex(new_starts, new_sizes)

        elif at == "end":
            starting_chunk = np.searchsorted(np.cumsum(self.__sizes), len(self) - n)
            new_sizes = self.__sizes[starting_chunk:].copy()
            new_starts = self.__starts[starting_chunk:].copy()
            new_sizes[0] = n - np.sum(new_sizes[1:])
            new_starts[0] = (
                self.__starts[starting_chunk]
                + self.__sizes[starting_chunk]
                - new_sizes[0]
            )
            return ChunkedIndex(new_starts, new_sizes)
        else:
            raise ValueError(f"Unknown value for 'at': {at}")

    def take_range(self, start: int, end: int) -> DataIndex:
        """
        Take a range of elements from the index.
        """
        if start < 0 or end > len(self):
            raise ValueError(
                f"Range {start}:{end} is out of bounds for index of size {len(self)}"
            )

        if start >= end:
            raise ValueError(f"Start {start} must be less than end {end}")

        # Get the indices of the chunks that are in the range
        idxs = np.concatenate(
            [
                np.arange(start, start + size)
                for start, size in zip(self.__starts, self.__sizes)
            ]
        )
        range_idxs = idxs[start:end]

        return SimpleIndex(range_idxs)

    def mask(self, mask: np.ndarray) -> DataIndex:
        """
        Mask the index with a boolean mask.
        """
        if mask.shape != (len(self),):
            raise ValueError(
                f"Mask shape {mask.shape} does not match index size {len(self)}"
            )

        if mask.dtype != bool:
            raise ValueError(f"Mask dtype {mask.dtype} is not boolean")

        if not mask.any():
            raise EmptyMaskError("Mask is all False")

        if mask.all():
            return self

        # Get the indices of the chunks that are masked
        idxs = np.concatenate(
            [
                np.arange(start, start + size)
                for start, size in zip(self.__starts, self.__sizes)
            ]
        )
        masked_idxs = idxs[mask]

        return SimpleIndex(masked_idxs)

    def get_data(self, data: h5py.Dataset | np.ndarray) -> np.ndarray:
        """
        Get the data from the dataset using the index. We want to perform as few reads
        as possible. However, the chunks may not be continuous. This method sorts the
        chunks so it can read the data in the largest possible chunks, it then
        reshuffles the data to match the original order.

        For large numbers of chunks, this is much much faster than reading each chunk
        in the order they are stored in the index. I know because I tried. It sucked.
        """
        if not isinstance(data, (h5py.Dataset, np.ndarray)):
            raise ValueError("Data must be a h5py.Dataset")

        if len(self) == 0:
            return np.array([], dtype=data.dtype)
        if len(self.__starts) == 1:
            return data[self.__starts[0] : self.__starts[0] + self.__sizes[0]]

        sorted_start_index = np.argsort(self.__starts)
        new_starts = self.__starts[sorted_start_index]
        new_sizes = self.__sizes[sorted_start_index]

        packed_starts, packed_sizes = pack(new_starts, new_sizes)

        shape = (len(self),) + data.shape[1:]
        temp = np.zeros(shape, dtype=data.dtype)
        running_index = 0
        for i, (start, size) in enumerate(zip(packed_starts, packed_sizes)):
            temp[running_index : running_index + size] = data[start : start + size]
            running_index += size

        output = np.zeros(len(self), dtype=data.dtype)
        cumulative_sorted_sizes = np.insert(np.cumsum(new_sizes), 0, 0)
        cumulative_original_sizes = np.insert(np.cumsum(self.__sizes), 0, 0)

        # reshuffle the output to match the original order
        for i, sorted_index in enumerate(sorted_start_index):
            start = cumulative_original_sizes[sorted_index]
            end = cumulative_original_sizes[sorted_index + 1]
            data = temp[cumulative_sorted_sizes[i] : cumulative_sorted_sizes[i + 1]]
            output[start:end] = data

        return output

    def __getitem__(self, item: int) -> DataIndex:
        """
        Get an item from the index.
        """
        if item < 0 or item >= len(self):
            raise IndexError(
                f"Index {item} out of bounds for index of size {len(self)}"
            )
        sums = np.cumsum(self.__sizes)
        index = np.searchsorted(sums, item)
        start = self.__starts[index]
        offset = item - sums[index - 1] if index > 0 else item
        return SimpleIndex(np.array([start + offset]))



from typing import Optional

from pydantic import BaseModel, model_validator


def empty_string_to_none(value: str) -> Optional[str]:
    if type(value) is str and value == "":
        return None
    return value


class CosmoToolsParameters(BaseModel):
    cosmotools_steps: list[int]
    fof_linking_length: float
    fof_pmin: int
    sod_pmin: int
    sod_delta_crit: float
    sod_concentration_pmin: int
    sodbighaloparticles_pmin: int
    profiles_nbins: int
    galaxy_dbscan_neighbors: Optional[int]
    galaxy_aperture_radius: Optional[int]
    galaxy_pmin: Optional[int]

    @model_validator(mode="before")
    @classmethod
    def empty_string_to_none(cls, data):
        if isinstance(data, dict):
            data = {k: empty_string_to_none(v) for k, v in data.items()}
        return data



from collections import defaultdict
from pathlib import Path
from typing import Iterable, Optional, Type

from h5py import File, Group

from opencosmo import dataset as d
from opencosmo import io
from opencosmo import link as l
from opencosmo.header import OpenCosmoHeader, read_header

try:
    from mpi4py import MPI

    from opencosmo.link.mpi import MpiLinkHandler
except ImportError:
    MPI = None  # type: ignore

LINK_ALIASES = {  # Left: Name in file, right: Name in collection
    "sodbighaloparticles_star_particles": "star_particles",
    "sodbighaloparticles_dm_particles": "dm_particles",
    "sodbighaloparticles_gravity_particles": "gravity_particles",
    "sodbighaloparticles_agn_particles": "agn_particles",
    "sodbighaloparticles_gas_particles": "gas_particles",
    "sod_profile": "halo_profiles",
    "galaxyproperties": "galaxy_properties",
    "galaxyparticles_star_particles": "star_particles",
}

ALLOWED_LINKS = {  # Files that can serve as a link holder and
    "halo_properties": ["halo_particles", "halo_profiles", "galaxy_properties"],
    "galaxy_properties": ["galaxy_particles"],
}


def verify_links(*headers: OpenCosmoHeader) -> tuple[str, list[str]]:
    """
    Verify that the links in the headers are valid. This means that the
    link holder has a corresponding link target and that the link target
    is of the correct type. It also verifies that the linked files are from
    the same simulation. Returns a dictionary where the keys are the
    link holder files and the values are lists of the corresponding link.

    Raises an error if the links are not valid, otherwise returns the links.
    """

    data_types = [header.file.data_type for header in headers]
    if len(set(data_types)) != len(data_types):
        raise ValueError("Data types in files must be unique to link correctly")

    properties_files = [dt for dt in data_types if dt in ALLOWED_LINKS]
    if not properties_files:
        raise ValueError("No valid link holder files found in headers")

    dtypes_to_headers = {header.file.data_type: header for header in headers}

    links = defaultdict(list)  # {file: [link_header, ...]}
    for file in properties_files:
        for link in ALLOWED_LINKS[file]:
            try:
                link_header = dtypes_to_headers[link]
                # Check that the headers come from the same simulation
                if link_header.simulation != dtypes_to_headers[file].simulation:
                    raise ValueError(f"Simulation mismatch between {file} and {link}")
                links[file].append(link)
            except KeyError:
                continue  # No link header found for this file

    has_links = [file in links for file in properties_files]
    # Properties files also need to have the same simulation
    if len(properties_files) > 1:
        # need exactly one true (for now)
        if sum(has_links) != 1:
            raise NotImplementedError("Chained links are not yet supported")
    for file in properties_files:
        if (
            dtypes_to_headers[file].simulation
            != dtypes_to_headers[properties_files[0]].simulation
        ):
            raise ValueError(
                f"Simulation mismatch between {file} and {properties_files[0]}"
            )
    properties_files = [
        file for file, has_link in zip(properties_files, has_links) if has_link
    ]
    property_file = properties_files[0]
    return property_file, links[property_file]


def open_linked_files(*files: Path):
    """
    Open a collection of files that are linked together, such as a
    properties file and a particle file.
    """
    file_handles = [File(file, "r") for file in files]
    headers = [read_header(file) for file in file_handles]
    properties_file, linked_files = verify_links(*headers)
    properties_index = next(
        index
        for index, header in enumerate(headers)
        if header.file.data_type == properties_file
    )
    properties_file = file_handles.pop(properties_index)
    properties_dataset = io.open(properties_file)
    if not isinstance(properties_dataset, d.Dataset):
        raise ValueError(
            "Properties file must contain a single dataset, but found more"
        )

    linked_files_by_type = {
        file["header"]["file"].attrs["data_type"]: file for file in file_handles
    }
    if len(linked_files_by_type) != len(linked_files):
        raise ValueError("Linked files must have unique data types")
    return get_linked_datasets(
        properties_dataset,
        linked_files_by_type,
        properties_file,
        headers[properties_index],
    )


def open_linked_file(
    file_handle: File, datasets_to_get: Optional[Iterable[str]] = None
) -> l.StructureCollection:
    """
    Open a single file that contains both properties and linked datasets.
    """
    properties_name = list(
        filter(lambda name: "properties" in name, file_handle.keys())
    )

    header = read_header(file_handle)
    if len(properties_name) == 2:
        if (
            "galaxy_properties" in properties_name
            and "halo_properties" in properties_name
        ):
            properties_name = ["halo_properties"]
            # Custom handling for now
        else:
            raise ValueError(
                "Multiple properties datasets found, please specify which one to use"
            )

    elif len(properties_name) == 0:
        raise ValueError("No properties dataset found in file")

    properties_name = properties_name[0]
    names_to_ignore = [properties_name, "header"] + list(datasets_to_get or [])
    other_datasets = [
        name for name in file_handle.keys() if name not in names_to_ignore
    ]
    if not other_datasets:
        raise ValueError("No linked datasets found in file")
    linked_groups_by_type = {name: file_handle[name] for name in other_datasets}
    properties_dataset = io.open(file_handle[properties_name])
    if not isinstance(properties_dataset, d.Dataset):
        raise ValueError("Properties dataset must be a single dataset")

    return get_linked_datasets(
        properties_dataset, linked_groups_by_type, file_handle[properties_name], header
    )


def get_linked_datasets(
    properties_dataset: d.Dataset,
    linked_files_by_type: dict[str, File | Group],
    properties_file: File,
    header: OpenCosmoHeader,
) -> l.StructureCollection:
    datasets = {}
    for dtype, pointer in linked_files_by_type.items():
        if "data" not in pointer.keys():
            datasets.update({k: pointer[k] for k in pointer.keys() if k != "header"})
        else:
            datasets.update({dtype: pointer})

    link_handlers = get_link_handlers(properties_file, datasets, header)
    output = {}
    for key, handler in link_handlers.items():
        if key in LINK_ALIASES:
            output[LINK_ALIASES[key]] = handler
        else:
            output[key] = handler

    return l.StructureCollection(properties_dataset, header, output)


def get_link_handlers(
    link_file: File | Group,
    linked_files: dict[str, File | Group],
    header: OpenCosmoHeader,
) -> dict[str, l.LinkHandler]:
    if "data_linked" not in link_file.keys():
        raise KeyError("No linked datasets found in the file.")
    links = link_file["data_linked"]

    handler: Type[l.LinkHandler]
    if MPI is not None and MPI.COMM_WORLD.Get_size() > 1:
        handler = MpiLinkHandler
    else:
        handler = l.OomLinkHandler
    unique_dtypes = {key.rsplit("_", 1)[0] for key in links.keys()}
    output_links = {}
    for dtype in unique_dtypes:
        if dtype not in linked_files and LINK_ALIASES.get(dtype) not in linked_files:
            continue  # Skip if the linked file is not provided

        key = LINK_ALIASES.get(dtype, dtype)
        if "data" not in linked_files[key].keys():
            raise KeyError(f"No data group found in linked file for dtype '{dtype}'")
        try:
            start = links[f"{dtype}_start"]
            size = links[f"{dtype}_size"]

            output_links[key] = handler(linked_files[key], (start, size), header)
        except KeyError:
            index = links[f"{dtype}_idx"]
            output_links[key] = handler(linked_files[key], index, header)
    return output_links
